"""
Base module for the tables
"""

import json

from epyk.core.html import Html

# The list of CSS classes
# from epyk.core.css.styles import CssGrpClsTable


class Row(Html.Html):
  name, category, callFnc = 'Row', 'Tables', None

  def __init__(self, report, cells):
    super(Row, self).__init__(report, cells)

  def __getitem__(self, i):
    return self.val[i]

  def cell(self, i):
    return self[i]

  def __str__(self):
    data = [v.html() for v in self.val]
    return "<tr %s>%s</tr>" % (self.get_attrs(pyClassNames=self.pyStyle), "".join(data))


class Cell(Html.Html):
  name, category, callFnc = 'Cell', 'Tables', None

  def __init__(self, report, text, is_header):
    super(Cell, self).__init__(report, text)
    self.is_header = is_header

  def set_html_content(self, htmlObj):
    """
    Set the cell content to be an HTML object

    :param htmlObj: Python HTML object
    :return: self, the cell object to allow the chaining
    """
    htmlObj.inReport = False
    self.innerPyHTML = htmlObj
    return self

  def __str__(self):
    if self.is_header:
      return "<th %s>%s</th>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.content)

    return "<td %s>%s</td>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.content)


class Bespoke(Html.Html):
  name, category, callFnc = 'Basic', 'Tables', 'basic'
  # _grpCls = CssGrpClsTable.CssClassTableBespoke

  def __init__(self, report, recordSet, cols, rows, width, height, htmlCode, options, profile):
    data = []
    self._fields = rows + cols
    for rec in recordSet:
      data.append([rec[c] for c in self._fields])
    super(Bespoke, self).__init__(report, data, code=htmlCode, width=width[0], widthUnit=width[1], height=height[0],
                                  heightUnit=height[1], profile=profile)
    self.items = None
    self.css({"text-align": 'center', 'border-collapse': 'collapse'})
    self._style = {"rows": {"padding": '5px 0'}, "header": {"padding": '5px 0'}}
    self.set_items()

  @property
  def tableId(self):
    """
    Return the Javascript variable of the bespoke
    """
    return self.dom.varId

  @property
  def header(self):
    """
    Get the header row. Returns none if missing
    """
    return self._header

  def set_items(self):
    if self.items is None:
      self.items = []
    if self._fields is not None:
      self._header = Row(self._report, [Cell(self._report, d, is_header=True) for d in self._fields])
      self.items.append(self._header)
    for rec in self.val:
      self.items.append(Row(self._report, [Cell(self._report, r, is_header=False) for r in rec]))
    return self

  def __getitem__(self, i):
    """
    Get the table rows
    """
    return self.items[i]

  def row(self, i, inc_header=False):
    """
    Get the table rows

    :param i: Integer. The column number
    :param inc_header: Boolean. Default False

    :return:
    """
    if not inc_header and self._fields is not None:
      return self[i+1]

    return self[i]

  def col(self, header=None, i=None):
    """
    Get the table column cells as a generator

    :param header: String.
    :param i: Integer
    """
    for v in self.items:
      if header is not None:
        i = self._fields.index(header)
      yield v[i]

    return self

  def add(self, row, missing="", is_header=False):
    """
    Add a row to the table

    Example
    simple_table.row_add({"column": "value"})

    :param data: The row to be added to the table
    :param missing: The data to put when a cell is missing

    :return: The python table
    """

    if isinstance(row, dict):
      data = [row.get(h, missing) for h in self._fields]
    else:
      data = row
    self.val.append(data)
    self.items.append(Row(self._report, [Cell(self._report, d, is_header=is_header) for d in data]))
    return self

  def __str__(self):
    str_rows = [r.html() for r in self.items]
    return "<table %s>%s</table>" % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(str_rows))


class Excel(Html.Html):
  name, category, callFnc = 'Excel', 'Tables', 'excel'
  # _grpCls = CssGrpClsTable.CssClassTableExcel

  def __init__(self, report, recordSet, cols, rows, title, width, height, cellwidth, delimiter, htmlCode):
    self.recordSet, self.delimiter = recordSet, delimiter
    super(Excel, self).__init__(report, [], code=htmlCode, width=width[0], widthUnit=width[1], height=height[0], heightUnit=height[1])
    self._jsStyles = {'header': rows + cols, 'cellwidth': cellwidth}
    self.css({'display': 'inline-block', 'overflow': 'auto', 'padding': 0, 'vertical-align': 'top'})
    self.add_title(title, options={'content_table': False})

  @property
  def tableId(self):
    """
    Return the Javascript variable of the bespoke
    """
    return self.dom.varId

  @property
  def val(self):
    return "JSON.stringify(tableData(%s))" % self.jqId

  @property
  def records(self):
    return "listToRec(tableData(%s), %s)" % (self.jqId, json.dumps(self._jsStyles['header']))

  @property
  def jqId(self):
    return "$('#%s table')" % self.htmlId

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, ''' htmlObj.empty();
      var tr = $('<tr></tr>');
      jsStyles.header.forEach(function(rec){tr.append("<th>"+ rec +"</th>")});
      htmlObj.append(tr); var tr = $('<tr></tr>'); var tbody = $('<tbody></tbody>');
      jsStyles.header.forEach(function(rec){tr.append("<td><input type='text' style='"+ jsStyles.cellwidth +"'/></td>")});
      tbody.append(tr);htmlObj.append(tbody)''')

  def __str__(self):
    self._report.jsOnLoadFnc.add('''
      function tableData(tableObj){
        res = [];
        tableObj.find('tbody').find('tr').each(function(key, val){
          var row = [];
          $(this).find('td').each(function(key, cell) {row.push($(cell).find('input').val())});
          res.push(row)}); return res};

      function listToRec(data, header){
          var res = [];
          data.forEach(function(row){
            rec = {};
            header.forEach(function(h, i){rec[h] = row[i]}); res.push(rec);
          }); return res}''')

    self.paste('''
      var tbody = $(this).find('tbody'); tbody.empty();
      var tableId = $(this).parent().attr('id');
      var lineDelimiter = $('#'+ tableId +'_delimiter').val();
      if (lineDelimiter == 'TAB'){lineDelimiter = '\\t'};
      data.split("\\n").forEach(function(line){
        if (line !== ''){
          var tr = $('<tr></tr>');
          line.split(lineDelimiter).forEach(function(rec){tr.append("<td><input type='text'  value='"+ rec +"'/></td>")
        }); tbody.append(tr)}}) ''')
    if self.delimiter is None:
      delimiter = '<input id="%s_delimiter" type="text" value="%s" placeholder="Line delimiter"/>' % (self.htmlId, self.delimiter)
    else:
      delimiter = '<input id="%s_delimiter" type="text" value="%s" style="display:none" placeholder="Line delimiter"/>' % (
          self.htmlId, self.delimiter)
    return '<div %(strAttr)s>%(delimiter)s<table style="width:100%%"></table></div>' % {
      'strAttr': self.get_attrs(pyClassNames=self.defined), 'delimiter': delimiter}



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_html_entities_4.asp
"""

AMPERSAND = '&amp;'
LESS_THAN = '&lt;'
GREATER_THAN = '&gt;'
NO_BREAK_SPACE = '&nbsp;'
NON_BREAKING_SPACE = '&nbsp;'
INVERTED_EXCLAMATION_MARK = '&iexcl;'
CENT_SIGN = '&cent;'
POUND_SIGN = '&pound;'
CURRENCY_SIGN = '&curren;'
YEN_SIGN = '&yen;'
YUAN_SIGN = '&yen;'
BROKEN_BAR = '&brvbar;'
BROKEN_VERTICAL_BAR = '&brvbar;'
SECTION_SIGN = '&sect;'
DIAERESIS = '&uml;'
SPACING_DIAERESIS = '&uml;'
COPYRIGHT_SIGN = '&copy;'
FEMININE_ORDINAL_INDICATOR = '&ordf;'
LEFT_POINTING_DOUBLE_ANGLE_QUOTATION_MARK = '&laquo;'
LEFT_POINTING_GUILLEMET = '&laquo;'
NOT_SIGN = '&not;'
SOFT_HYPHEN = '&shy;'
DISCRETIONARY_HYPHEN = '&shy;'
REGISTERED_SIGN = '&reg;'
REGISTERED_TRADE_MARK_SIGN = '&reg;'
MACRON = '&macr;'
SPACING_MACRON = '&macr;'
OVERLINE = '&macr;'
APL_OVERBAR = '&macr;'
DEGREE_SIGN = '&deg;'
PLUS_MINUS_SIGN = '&plusmn;'
PLUS_OR_MINUS_SIGN = '&plusmn;'
SUPERSCRIPT_TWO = '&sup2;'
SUPERSCRIPT_DIGIT_TWO = '&sup2;'
SQUARED = '&sup2;'
SUPERSCRIPT_THREE = '&sup3;'
SUPERSCRIPT_DIGIT_THREE = '&sup3;'
CUBED = '&sup3;'
ACUTE_ACCENT = '&acute;'
SPACING_ACUTE = '&acute;'
MICRO_SIGN = '&micro;'
PILCROW_SIGN = '&para;'
PARAGRAPH_SIGN = '&para;'
MIDDLE_DOT = '&middot;'
GEORGIAN_COMMA = '&middot;'
GREEK_MIDDLE_DOT = '&middot;'
CEDILLA = '&cedil;'
SPACING_CEDILLA = '&cedil;'
SUPERSCRIPT_ONE = '&sup1;'
SUPERSCRIPT_DIGIT_ONE = '&sup1;'
MASCULINE_ORDINAL_INDICATOR = '&ordm;'
RIGHT_POINTING_DOUBLE_ANGLE_QUOTATION_MARK = '&raquo;'
RIGHT_POINTING_GUILLEMET = '&raquo;'
VULGAR_FRACTION_ONE_QUARTER = '&frac14;'
FRACTION_ONE_QUARTER = '&frac14;'
VULGAR_FRACTION_ONE_HALF = '&frac12;'
FRACTION_ONE_HALF = '&frac12;'
VULGAR_FRACTION_THREE_QUARTERS = '&frac34;'
FRACTION_THREE_QUARTERS = '&frac34;'
INVERTED_QUESTION_MARK = '&iquest;'
TURNED_QUESTION_MARK = '&iquest;'
LATIN_CAPITAL_LETTER_A_WITH_GRAVE = '&Agrave;'
LATIN_CAPITAL_LETTER_A_GRAVE = '&Agrave;'
LATIN_CAPITAL_LETTER_A_WITH_ACUTE = '&Aacute;'
LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX = '&Acirc;'
LATIN_CAPITAL_LETTER_A_WITH_TILDE = '&Atilde;'
LATIN_CAPITAL_LETTER_A_WITH_DIAERESIS = '&Auml;'
LATIN_CAPITAL_LETTER_A_WITH_RING_ABOVE = '&Aring;'
LATIN_CAPITAL_LETTER_A_RING = '&Aring;'
LATIN_CAPITAL_LETTER_AE = '&AElig;'
LATIN_CAPITAL_LIGATURE_AE = '&AElig;'
LATIN_CAPITAL_LETTER_C_WITH_CEDILLA = '&Ccedil;'
LATIN_CAPITAL_LETTER_E_WITH_GRAVE = '&Egrave;'
LATIN_CAPITAL_LETTER_E_WITH_ACUTE = '&Eacute;'
LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX = '&Ecirc;'
LATIN_CAPITAL_LETTER_E_WITH_DIAERESIS = '&Euml;'
LATIN_CAPITAL_LETTER_I_WITH_GRAVE = '&Igrave;'
LATIN_CAPITAL_LETTER_I_WITH_ACUTE = '&Iacute;'
LATIN_CAPITAL_LETTER_I_WITH_CIRCUMFLEX = '&Icirc;'
LATIN_CAPITAL_LETTER_I_WITH_DIAERESIS = '&Iuml;'
LATIN_CAPITAL_LETTER_ETH = '&ETH;'
LATIN_CAPITAL_LETTER_N_WITH_TILDE = '&Ntilde;'
LATIN_CAPITAL_LETTER_O_WITH_GRAVE = '&Ograve;'
LATIN_CAPITAL_LETTER_O_WITH_ACUTE = '&Oacute;'
LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX = '&Ocirc;'
LATIN_CAPITAL_LETTER_O_WITH_TILDE = '&Otilde;'
LATIN_CAPITAL_LETTER_O_WITH_DIAERESIS = '&Ouml;'
MULTIPLICATION_SIGN = '&times;'
LATIN_CAPITAL_LETTER_O_WITH_STROKE = '&Oslash;'
LATIN_CAPITAL_LETTER_O_SLASH = '&Oslash;'
LATIN_CAPITAL_LETTER_U_WITH_GRAVE = '&Ugrave;'
LATIN_CAPITAL_LETTER_U_WITH_ACUTE = '&Uacute;'
LATIN_CAPITAL_LETTER_U_WITH_CIRCUMFLEX = '&Ucirc;'
LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS = '&Uuml;'
LATIN_CAPITAL_LETTER_Y_WITH_ACUTE = '&Yacute;'
LATIN_CAPITAL_LETTER_THORN = '&THORN;'
LATIN_SMALL_LETTER_SHARP_S = '&szlig;'
ESS_ZED = '&szlig;'
LATIN_SMALL_LETTER_A_WITH_GRAVE = '&agrave;'
LATIN_SMALL_LETTER_A_GRAVE = '&agrave;'
LATIN_SMALL_LETTER_A_WITH_ACUTE = '&aacute;'
LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX = '&acirc;'
LATIN_SMALL_LETTER_A_WITH_TILDE = '&atilde;'
LATIN_SMALL_LETTER_A_WITH_DIAERESIS = '&auml;'
LATIN_SMALL_LETTER_A_WITH_RING_ABOVE = '&aring;'
LATIN_SMALL_LETTER_A_RING = '&aring;'
LATIN_SMALL_LETTER_AE = '&aelig;'
LATIN_SMALL_LIGATURE_AE = '&aelig;'
LATIN_SMALL_LETTER_C_WITH_CEDILLA = '&ccedil;'
LATIN_SMALL_LETTER_E_WITH_GRAVE = '&egrave;'
LATIN_SMALL_LETTER_E_WITH_ACUTE = '&eacute;'
LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX = '&ecirc;'
LATIN_SMALL_LETTER_E_WITH_DIAERESIS = '&euml;'
LATIN_SMALL_LETTER_I_WITH_GRAVE = '&igrave;'
LATIN_SMALL_LETTER_I_WITH_ACUTE = '&iacute;'
LATIN_SMALL_LETTER_I_WITH_CIRCUMFLEX = '&icirc;'
LATIN_SMALL_LETTER_I_WITH_DIAERESIS = '&iuml;'
LATIN_SMALL_LETTER_ETH = '&eth;'
LATIN_SMALL_LETTER_N_WITH_TILDE = '&ntilde;'
LATIN_SMALL_LETTER_O_WITH_GRAVE = '&ograve;'
LATIN_SMALL_LETTER_O_WITH_ACUTE = '&oacute;'
LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX = '&ocirc;'
LATIN_SMALL_LETTER_O_WITH_TILDE = '&otilde;'
LATIN_SMALL_LETTER_O_WITH_DIAERESIS = '&ouml;'
DIVISION_SIGN = '&divide;'
LATIN_SMALL_LETTER_O_WITH_STROKE = '&oslash;'
LATIN_SMALL_LETTER_O_SLASH = '&oslash;'
LATIN_SMALL_LETTER_U_WITH_GRAVE = '&ugrave;'
LATIN_SMALL_LETTER_U_WITH_ACUTE = '&uacute;'
LATIN_SMALL_LETTER_U_WITH_CIRCUMFLEX = '&ucirc;'
LATIN_SMALL_LETTER_U_WITH_DIAERESIS = '&uuml;'
LATIN_SMALL_LETTER_Y_WITH_ACUTE = '&yacute;'
LATIN_SMALL_LETTER_THORN = '&thorn;'
LATIN_SMALL_LETTER_Y_WITH_DIAERESIS = '&yuml;'
LATIN_SMALL_F_WITH_HOOK = '&fnof;'
FUNCTION = '&fnof;'
FLORIN = '&fnof;'
GREEK_CAPITAL_LETTER_ALPHA = '&Alpha;'
GREEK_CAPITAL_LETTER_BETA = '&Beta;'
GREEK_CAPITAL_LETTER_GAMMA = '&Gamma;'
GREEK_CAPITAL_LETTER_DELTA = '&Delta;'
GREEK_CAPITAL_LETTER_EPSILON = '&Epsilon;'
GREEK_CAPITAL_LETTER_ZETA = '&Zeta;'
GREEK_CAPITAL_LETTER_ETA = '&Eta;'
GREEK_CAPITAL_LETTER_THETA = '&Theta;'
GREEK_CAPITAL_LETTER_IOTA = '&Iota;'
GREEK_CAPITAL_LETTER_KAPPA = '&Kappa;'
GREEK_CAPITAL_LETTER_LAMBDA = '&Lambda;'
GREEK_CAPITAL_LETTER_MU = '&Mu;'
GREEK_CAPITAL_LETTER_NU = '&Nu;'
GREEK_CAPITAL_LETTER_XI = '&Xi;'
GREEK_CAPITAL_LETTER_OMICRON = '&Omicron;'
GREEK_CAPITAL_LETTER_PI = '&Pi;'
GREEK_CAPITAL_LETTER_RHO = '&Rho;'
GREEK_CAPITAL_LETTER_SIGMA = '&Sigma;'
GREEK_CAPITAL_LETTER_TAU = '&Tau;'
GREEK_CAPITAL_LETTER_UPSILON = '&Upsilon;'
GREEK_CAPITAL_LETTER_PHI = '&Phi;'
GREEK_CAPITAL_LETTER_CHI = '&Chi;'
GREEK_CAPITAL_LETTER_PSI = '&Psi;'
GREEK_CAPITAL_LETTER_OMEGA = '&Omega;'
GREEK_SMAL_LETTER_ALPHA = '&alpha;'
GREEK_SMAL_LETTER_BETA = '&beta;'
GREEK_SMAL_LETTER_GAMMA = '&gamma;'
GREEK_SMAL_LETTER_DELTA = '&delta;'
GREEK_SMAL_LETTER_EPSILON = '&epsilon;'
GREEK_SMAL_LETTER_ZETA = '&zeta;'
GREEK_SMAL_LETTER_ETA = '&eta;'
GREEK_SMAL_LETTER_THETA = '&theta;'
GREEK_SMAL_LETTER_IOTA = '&iota;'
GREEK_SMAL_LETTER_KAPPA = '&kappa;'
GREEK_SMAL_LETTER_LAMBDA = '&lambda;'
GREEK_SMAL_LETTER_MU = '&mu;'
GREEK_SMAL_LETTER_NU = '&nu;'
GREEK_SMAL_LETTER_XI = '&xi;'
GREEK_SMAL_LETTER_OMICRON = '&omicron;'
GREEK_SMAL_LETTER_PI = '&pi;'
GREEK_SMAL_LETTER_RHO = '&rho;'
GREEK_SMAL_LETTER_FINAL_SIGMA = '&sigmaf;'
GREEK_SMAL_LETTER_SIGMA = '&sigma;'
GREEK_SMAL_LETTER_TAU = '&tau;'
GREEK_SMAL_LETTER_UPSILON = '&upsilon;'
GREEK_SMAL_LETTER_PHI = '&phi;'
GREEK_SMAL_LETTER_CHI = '&chi;'
GREEK_SMAL_LETTER_PSI = '&psi;'
GREEK_SMAL_LETTER_OMEGA = '&omega;'
GREEK_SMAL_LETTER_THETA_SYMBOL = '&thetasym;'
GREEK_UPSILON_WITH_HOOK_SYMBOL = '&upsih;'
GREEK_PI_SYMBOL = '&piv;'




from epyk.core.html import graph


class C3(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "C3"

  def line(self, record, y_columns=None, x_axis=None, title=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def spline(self, record, y_columns=None, x_axis=None, title=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def step(self, record, y_columns=None, x_axis=None, title=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = 'step'
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def area(self, record, y_columns=None, x_axis=None, title=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def area_step(self, record, y_columns=None, x_axis=None, title=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = "area-step"
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def timeseries(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
                 width=(100, "%"), height=(330, "px"), htmlCode=None):
    line = self.line(record, y_columns, x_axis, title, profile, options, width, height, htmlCode)
    line.axis.x.type = "timeseries"
    line.axis.x.tick.format = "%Y-%m-%d"
    return line

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def hbar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    https://c3js.org/samples/axes_rotated.html

    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    h_bar = self.bar(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    h_bar.axis.rotated = True
    return h_bar

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    line_chart = graph.GraphC3.ChartScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, data[i])
    self.parent.context.register(line_chart)
    return line_chart

  def pie(self, record, y_column=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    https://c3js.org/samples/chart_pie.html

    :param record:
    :param y_column:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      if y_column in rec:
        agg_data[rec[x_axis]] = agg_data.get(rec[x_axis],  0) + float(rec[y_column])

    pie_chart = graph.GraphC3.ChartPie(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for k, value in agg_data.items():
      pie_chart.add_dataset(k, value)
    self.parent.context.register(pie_chart)
    return pie_chart

  def donut(self, record, y_column=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    agg_data = {}
    for rec in record:
      if y_column in rec:
        agg_data[rec[x_axis]] = agg_data.get(rec[x_axis],  0) + float(rec[y_column])

    pie_chart = graph.GraphC3.ChartDonut(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for k, value in agg_data.items():
      pie_chart.add_dataset(k, value)
    self.parent.context.register(pie_chart)
    return pie_chart

  def gauge(self, value, text="", profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param value:
    :param text:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    g_chart = graph.GraphC3.ChartGauge(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    g_chart.add_dataset(text, value)
    self.parent.context.register(g_chart)
    return g_chart

  def stanford(self, record, y_columns=None, x_axis=None, epoch_col=None, title=None, profile=None, width=(100, "%"),
               height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param y_columns:
    :param x_axis:
    :param epoch_col:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    epoch, labels, series = [], [], []
    for rec in record:
      epoch.append(rec[epoch_col])
      labels.append(rec[x_axis])
      series.append([rec.get(c)for c in y_columns])

    line_chart = graph.GraphC3.ChartStanford(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(labels)
    line_chart.epoch(epoch, epoch_col)
    for i, y in enumerate(y_columns):
      line_chart.add_dataset(y, series[i])
    self.parent.context.register(line_chart)
    return line_chart



"""
Related Pages:

		https://www.w3schools.com/icons/fontawesome_icons_webapp.asp
"""

ICON_ADDRESS_BOOK = 'fa fa-address-book'
ICON_ADDRESS_BOOK_O = 'fa fa-address-book-o'
ICON_ADDRESS_CARD = 'fa fa-address-card'
ICON_ADDRESS_CARD_O = 'fa fa-address-card-o'
ICON_ADJUST = 'fa fa-adjust'
ICON_AMERICAN_SIGN_LANGUAGE_INTERPRETING = 'fa fa-american-sign-language-interpreting'
ICON_ANCHOR = 'fa fa-anchor'
ICON_ARCHIVE = 'fa fa-archive'
ICON_AREA_CHART = 'fa fa-area-chart'
ICON_ARROWS = 'fa fa-arrows'
ICON_ARROWS_H = 'fa fa-arrows-h'
ICON_ARROWS_V = 'fa fa-arrows-v'
ICON_ASL_INTERPRETING = 'fa fa-asl-interpreting'
ICON_ASSISTIVE_LISTENING_SYSTEMS = 'fa fa-assistive-listening-systems'
ICON_ASTERISK = 'fa fa-asterisk'
ICON_AT = 'fa fa-at'
ICON_AUTOMOBILE = 'fa fa-automobile'
ICON_AUDIO_DESCRIPTION = 'fa fa-audio-description'
ICON_BALANCE_SCALE = 'fa fa-balance-scale'
ICON_BAN = 'fa fa-ban'
ICON_BANK = 'fa fa-bank'
ICON_BAR_CHART = 'fa fa-bar-chart'
ICON_BAR_CHART_O = 'fa fa-bar-chart-o'
ICON_BARCODE = 'fa fa-barcode'
ICON_BARS = 'fa fa-bars'
ICON_BATH = 'fa fa-bath'
ICON_BATHTUB = 'fa fa-bathtub'
ICON_BATTERY_0 = 'fa fa-battery-0'
ICON_BATTERY_1 = 'fa fa-battery-1'
ICON_BATTERY_2 = 'fa fa-battery-2'
ICON_BATTERY_3 = 'fa fa-battery-3'
ICON_BATTERY_4 = 'fa fa-battery-4'
ICON_BATTERY_EMPTY = 'fa fa-battery-empty'
ICON_BATTERY_FULL = 'fa fa-battery-full'
ICON_BATTERY_HALF = 'fa fa-battery-half'
ICON_BATTERY_QUARTER = 'fa fa-battery-quarter'
ICON_BATTERY_THREE_QUARTERS = 'fa fa-battery-three-quarters'
ICON_BED = 'fa fa-bed'
ICON_BEER = 'fa fa-beer'
ICON_BELL = 'fa fa-bell'
ICON_BELL_O = 'fa fa-bell-o'
ICON_BELL_SLASH = 'fa fa-bell-slash'
ICON_BELL_SLASH_O = 'fa fa-bell-slash-o'
ICON_BICYCLE = 'fa fa-bicycle'
ICON_BINOCULARS = 'fa fa-binoculars'
ICON_BIRTHDAY_CAKE = 'fa fa-birthday-cake'
ICON_BLIND = 'fa fa-blind'
ICON_BOLT = 'fa fa-bolt'
ICON_BOMB = 'fa fa-bomb'
ICON_BOOK = 'fa fa-book'
ICON_BOOKMARK = 'fa fa-bookmark'
ICON_BOOKMARK_O = 'fa fa-bookmark-o'
ICON_BRAILLE = 'fa fa-braille'
ICON_BRIEFCASE = 'fa fa-briefcase'
ICON_BUG = 'fa fa-bug'
ICON_BUILDING = 'fa fa-building'
ICON_BUILDING_O = 'fa fa-building-o'
ICON_BULLHORN = 'fa fa-bullhorn'
ICON_BULLSEYE = 'fa fa-bullseye'
ICON_BUS = 'fa fa-bus'
ICON_CAB = 'fa fa-cab'
ICON_CALCULATOR = 'fa fa-calculator'
ICON_CALENDAR = 'fa fa-calendar'
ICON_CALENDAR_O = 'fa fa-calendar-o'
ICON_CALENDAR_CHECK_O = 'fa fa-calendar-check-o'
ICON_CALENDAR_MINUS_O = 'fa fa-calendar-minus-o'
ICON_CALENDAR_PLUS_O = 'fa fa-calendar-plus-o'
ICON_CALENDAR_TIMES_O = 'fa fa-calendar-times-o'
ICON_CAMERA = 'fa fa-camera'
ICON_CAMERA_RETRO = 'fa fa-camera-retro'
ICON_CAR = 'fa fa-car'
ICON_CARET_SQUARE_O_DOWN = 'fa fa-caret-square-o-down'
ICON_CARET_SQUARE_O_LEFT = 'fa fa-caret-square-o-left'
ICON_CARET_SQUARE_O_RIGHT = 'fa fa-caret-square-o-right'
ICON_CARET_SQUARE_O_UP = 'fa fa-caret-square-o-up'
ICON_CART_ARROW_DOWN = 'fa fa-cart-arrow-down'
ICON_CART_PLUS = 'fa fa-cart-plus'
ICON_CC = 'fa fa-cc'
ICON_CERTIFICATE = 'fa fa-certificate'
ICON_CHECK = 'fa fa-check'
ICON_CHECK_CIRCLE = 'fa fa-check-circle'
ICON_CHECK_CIRCLE_O = 'fa fa-check-circle-o'
ICON_CHECK_SQUARE = 'fa fa-check-square'
ICON_CHECK_SQUARE_O = 'fa fa-check-square-o'
ICON_CHILD = 'fa fa-child'
ICON_CIRCLE = 'fa fa-circle'
ICON_CIRCLE_O = 'fa fa-circle-o'
ICON_CIRCLE_O_NOTCH = 'fa fa-circle-o-notch'
ICON_CIRCLE_THIN = 'fa fa-circle-thin'
ICON_CLOCK_O = 'fa fa-clock-o'
ICON_CLONE = 'fa fa-clone'
ICON_CLOSE = 'fa fa-close'
ICON_CLOUD = 'fa fa-cloud'
ICON_CLOUD_DOWNLOAD = 'fa fa-cloud-download'
ICON_CLOUD_UPLOAD = 'fa fa-cloud-upload'
ICON_CODE = 'fa fa-code'
ICON_CODE_FORK = 'fa fa-code-fork'
ICON_COFFEE = 'fa fa-coffee'
ICON_COG = 'fa fa-cog'
ICON_COGS = 'fa fa-cogs'
ICON_COMMENT = 'fa fa-comment'
ICON_COMMENT_O = 'fa fa-comment-o'
ICON_COMMENTS = 'fa fa-comments'
ICON_COMMENTS_O = 'fa fa-comments-o'
ICON_COMMENTING = 'fa fa-commenting'
ICON_COMMENTING_O = 'fa fa-commenting-o'
ICON_COMPASS = 'fa fa-compass'
ICON_COPYRIGHT = 'fa fa-copyright'
ICON_CREDIT_CARD = 'fa fa-credit-card'
ICON_CREDIT_CARD_ALT = 'fa fa-credit-card-alt'
ICON_CREATIVE_COMMONS = 'fa fa-creative-commons'
ICON_CROP = 'fa fa-crop'
ICON_CROSSHAIRS = 'fa fa-crosshairs'
ICON_CUBE = 'fa fa-cube'
ICON_CUBES = 'fa fa-cubes'
ICON_CUTLERY = 'fa fa-cutlery'
ICON_DASHBOARD = 'fa fa-dashboard'
ICON_DATABASE = 'fa fa-database'
ICON_DEAF = 'fa fa-deaf'
ICON_DEAFNESS = 'fa fa-deafness'
ICON_DESKTOP = 'fa fa-desktop'
ICON_DIAMOND = 'fa fa-diamond'
ICON_DOT_CIRCLE_O = 'fa fa-dot-circle-o'
ICON_DOWNLOAD = 'fa fa-download'
ICON_DRIVERS_LICENSE = 'fa fa-drivers-license'
ICON_DRIVERS_LICENSE_O = 'fa fa-drivers-license-o'
ICON_EDIT = 'fa fa-edit'
ICON_ELLIPSIS_H = 'fa fa-ellipsis-h'
ICON_ELLIPSIS_V = 'fa fa-ellipsis-v'
ICON_ENVELOPE = 'fa fa-envelope'
ICON_ENVELOPE_O = 'fa fa-envelope-o'
ICON_ENVELOPE_OPEN = 'fa fa-envelope-open'
ICON_ENVELOPE_OPEN_O = 'fa fa-envelope-open-o'
ICON_ENVELOPE_SQUARE = 'fa fa-envelope-square'
ICON_ERASER = 'fa fa-eraser'
ICON_EXCHANGE = 'fa fa-exchange'
ICON_EXCLAMATION = 'fa fa-exclamation'
ICON_EXCLAMATION_CIRCLE = 'fa fa-exclamation-circle'
ICON_EXCLAMATION_TRIANGLE = 'fa fa-exclamation-triangle'
ICON_EXTERNAL_LINK = 'fa fa-external-link'
ICON_EXTERNAL_LINK_SQUARE = 'fa fa-external-link-square'
ICON_EYE = 'fa fa-eye'
ICON_EYE_SLASH = 'fa fa-eye-slash'
ICON_EYEDROPPER = 'fa fa-eyedropper'
ICON_FAX = 'fa fa-fax'
ICON_FEMALE = 'fa fa-female'
ICON_FIGHTER_JET = 'fa fa-fighter-jet'
ICON_FILE_ARCHIVE_O = 'fa fa-file-archive-o'
ICON_FILE_AUDIO_O = 'fa fa-file-audio-o'
ICON_FILE_CODE_O = 'fa fa-file-code-o'
ICON_FILE_EXCEL_O = 'fa fa-file-excel-o'
ICON_FILE_IMAGE_O = 'fa fa-file-image-o'
ICON_FILE_MOVIE_O = 'fa fa-file-movie-o'
ICON_FILE_PDF_O = 'fa fa-file-pdf-o'
ICON_FILE_PHOTO_O = 'fa fa-file-photo-o'
ICON_FILE_PICTURE_O = 'fa fa-file-picture-o'
ICON_FILE_POWERPOINT_O = 'fa fa-file-powerpoint-o'
ICON_FILE_SOUND_O = 'fa fa-file-sound-o'
ICON_FILE_VIDEO_O = 'fa fa-file-video-o'
ICON_FILE_WORD_O = 'fa fa-file-word-o'
ICON_FILE_ZIP_O = 'fa fa-file-zip-o'
ICON_FILM = 'fa fa-film'
ICON_FILTER = 'fa fa-filter'
ICON_FIRE = 'fa fa-fire'
ICON_FIRE_EXTINGUISHER = 'fa fa-fire-extinguisher'
ICON_FLAG = 'fa fa-flag'
ICON_FLAG_CHECKERED = 'fa fa-flag-checkered'
ICON_FLAG_O = 'fa fa-flag-o'
ICON_FLASH = 'fa fa-flash'
ICON_FLASK = 'fa fa-flask'
ICON_FOLDER = 'fa fa-folder'
ICON_FOLDER_O = 'fa fa-folder-o'
ICON_FOLDER_OPEN = 'fa fa-folder-open'
ICON_FOLDER_OPEN_O = 'fa fa-folder-open-o'
ICON_FROWN_O = 'fa fa-frown-o'
ICON_FUTBOL_O = 'fa fa-futbol-o'
ICON_GAMEPAD = 'fa fa-gamepad'
ICON_GAVEL = 'fa fa-gavel'
ICON_GEAR = 'fa fa-gear'
ICON_GEARS = 'fa fa-gears'
ICON_GENDERLESS = 'fa fa-genderless'
ICON_GIFT = 'fa fa-gift'
ICON_GLASS = 'fa fa-glass'
ICON_GLOBE = 'fa fa-globe'
ICON_GRADUATION_CAP = 'fa fa-graduation-cap'
ICON_GROUP = 'fa fa-group'
ICON_HARD_OF_HEARING = 'fa fa-hard-of-hearing'
ICON_HDD_O = 'fa fa-hdd-o'
ICON_HANDSHAKE_O = 'fa fa-handshake-o'
ICON_HASHTAG = 'fa fa-hashtag'
ICON_HEADPHONES = 'fa fa-headphones'
ICON_HEART = 'fa fa-heart'
ICON_HEART_O = 'fa fa-heart-o'
ICON_HEARTBEAT = 'fa fa-heartbeat'
ICON_HISTORY = 'fa fa-history'
ICON_HOME = 'fa fa-home'
ICON_HOTEL = 'fa fa-hotel'
ICON_HOURGLASS = 'fa fa-hourglass'
ICON_HOURGLASS_1 = 'fa fa-hourglass-1'
ICON_HOURGLASS_2 = 'fa fa-hourglass-2'
ICON_HOURGLASS_3 = 'fa fa-hourglass-3'
ICON_HOURGLASS_END = 'fa fa-hourglass-end'
ICON_HOURGLASS_HALF = 'fa fa-hourglass-half'
ICON_HOURGLASS_O = 'fa fa-hourglass-o'
ICON_HOURGLASS_START = 'fa fa-hourglass-start'
ICON_I_CURSOR = 'fa fa-i-cursor'
ICON_ID_BADGE = 'fa fa-id-badge'
ICON_ID_CARD = 'fa fa-id-card'
ICON_ID_CARD_O = 'fa fa-id-card-o'
ICON_IMAGE = 'fa fa-image'
ICON_INBOX = 'fa fa-inbox'
ICON_INDUSTRY = 'fa fa-industry'
ICON_INFO = 'fa fa-info'
ICON_INFO_CIRCLE = 'fa fa-info-circle'
ICON_INSTITUTION = 'fa fa-institution'
ICON_KEY = 'fa fa-key'
ICON_KEYBOARD_O = 'fa fa-keyboard-o'
ICON_LANGUAGE = 'fa fa-language'
ICON_LAPTOP = 'fa fa-laptop'
ICON_LEAF = 'fa fa-leaf'
ICON_LEGAL = 'fa fa-legal'
ICON_LEMON_O = 'fa fa-lemon-o'
ICON_LEVEL_DOWN = 'fa fa-level-down'
ICON_LEVEL_UP = 'fa fa-level-up'
ICON_LIFE_BOUY = 'fa fa-life-bouy'
ICON_LIFE_BUOY = 'fa fa-life-buoy'
ICON_LIFE_RING = 'fa fa-life-ring'
ICON_LIFE_SAVER = 'fa fa-life-saver'
ICON_LIGHTBULB_O = 'fa fa-lightbulb-o'
ICON_LINE_CHART = 'fa fa-line-chart'
ICON_LOCATION_ARROW = 'fa fa-location-arrow'
ICON_LOCK = 'fa fa-lock'
ICON_LOW_VISION = 'fa fa-low-vision'
ICON_MAGIC = 'fa fa-magic'
ICON_MAGNET = 'fa fa-magnet'
ICON_MAIL_FORWARD = 'fa fa-mail-forward'
ICON_MAIL_REPLY = 'fa fa-mail-reply'
ICON_MAIL_REPLY_ALL = 'fa fa-mail-reply-all'
ICON_MALE = 'fa fa-male'
ICON_MAP = 'fa fa-map'
ICON_MAP_O = 'fa fa-map-o'
ICON_MAP_PIN = 'fa fa-map-pin'
ICON_MAP_SIGNS = 'fa fa-map-signs'
ICON_MAP_MARKER = 'fa fa-map-marker'
ICON_MEH_O = 'fa fa-meh-o'
ICON_MICROCHIP = 'fa fa-microchip'
ICON_MICROPHONE = 'fa fa-microphone'
ICON_MICROPHONE_SLASH = 'fa fa-microphone-slash'
ICON_MINUS = 'fa fa-minus'
ICON_MINUS_CIRCLE = 'fa fa-minus-circle'
ICON_MINUS_SQUARE = 'fa fa-minus-square'
ICON_MINUS_SQUARE_O = 'fa fa-minus-square-o'
ICON_MOBILE = 'fa fa-mobile'
ICON_MOBILE_PHONE = 'fa fa-mobile-phone'
ICON_MONEY = 'fa fa-money'
ICON_MOON_O = 'fa fa-moon-o'
ICON_MORTAR_BOARD = 'fa fa-mortar-board'
ICON_MOTORCYCLE = 'fa fa-motorcycle'
ICON_MOUSE_POINTER = 'fa fa-mouse-pointer'
ICON_MUSIC = 'fa fa-music'
ICON_NAVICON = 'fa fa-navicon'
ICON_NEWSPAPER_O = 'fa fa-newspaper-o'
ICON_OBJECT_GROUP = 'fa fa-object-group'
ICON_OBJECT_UNGROUP = 'fa fa-object-ungroup'
ICON_PAINT_BRUSH = 'fa fa-paint-brush'
ICON_PAPER_PLANE = 'fa fa-paper-plane'
ICON_PAPER_PLANE_O = 'fa fa-paper-plane-o'
ICON_PAW = 'fa fa-paw'
ICON_PENCIL = 'fa fa-pencil'
ICON_PENCIL_SQUARE = 'fa fa-pencil-square'
ICON_PENCIL_SQUARE_O = 'fa fa-pencil-square-o'
ICON_PERCENT = 'fa fa-percent'
ICON_PHONE = 'fa fa-phone'
ICON_PHONE_SQUARE = 'fa fa-phone-square'
ICON_PHOTO = 'fa fa-photo'
ICON_PICTURE_O = 'fa fa-picture-o'
ICON_PIE_CHART = 'fa fa-pie-chart'
ICON_PLANE = 'fa fa-plane'
ICON_PLUG = 'fa fa-plug'
ICON_PLUS = 'fa fa-plus'
ICON_PLUS_CIRCLE = 'fa fa-plus-circle'
ICON_PLUS_SQUARE = 'fa fa-plus-square'
ICON_PLUS_SQUARE_O = 'fa fa-plus-square-o'
ICON_PODCAST = 'fa fa-podcast'
ICON_POWER_OFF = 'fa fa-power-off'
ICON_PRINT = 'fa fa-print'
ICON_PUZZLE_PIECE = 'fa fa-puzzle-piece'
ICON_QRCODE = 'fa fa-qrcode'
ICON_QUESTION = 'fa fa-question'
ICON_QUESTION_CIRCLE = 'fa fa-question-circle'
ICON_QUESTION_CIRCLE_O = 'fa fa-question-circle-o'
ICON_QUOTE_LEFT = 'fa fa-quote-left'
ICON_QUOTE_RIGHT = 'fa fa-quote-right'
ICON_RANDOM = 'fa fa-random'
ICON_RECYCLE = 'fa fa-recycle'
ICON_REFRESH = 'fa fa-refresh'
ICON_REGISTERED = 'fa fa-registered'
ICON_REMOVE = 'fa fa-remove'
ICON_REORDER = 'fa fa-reorder'
ICON_REPLY = 'fa fa-reply'
ICON_REPLY_ALL = 'fa fa-reply-all'
ICON_RETWEET = 'fa fa-retweet'
ICON_ROAD = 'fa fa-road'
ICON_ROCKET = 'fa fa-rocket'
ICON_RSS = 'fa fa-rss'
ICON_RSS_SQUARE = 'fa fa-rss-square'
ICON_S15 = 'fa fa-s15'
ICON_SEARCH = 'fa fa-search'
ICON_SEARCH_MINUS = 'fa fa-search-minus'
ICON_SEARCH_PLUS = 'fa fa-search-plus'
ICON_SEND = 'fa fa-send'
ICON_SEND_O = 'fa fa-send-o'
ICON_SERVER = 'fa fa-server'
ICON_SHARE = 'fa fa-share'
ICON_SHARE_ALT = 'fa fa-share-alt'
ICON_SHARE_ALT_SQUARE = 'fa fa-share-alt-square'
ICON_SHARE_SQUARE = 'fa fa-share-square'
ICON_SHARE_SQUARE_O = 'fa fa-share-square-o'
ICON_SHIELD = 'fa fa-shield'
ICON_SHIP = 'fa fa-ship'
ICON_SHOPPING_BAG = 'fa fa-shopping-bag'
ICON_SHOPPING_BASKET = 'fa fa-shopping-basket'
ICON_SHOPPING_CART = 'fa fa-shopping-cart'
ICON_SHOWER = 'fa fa-shower'
ICON_SIGN_IN = 'fa fa-sign-in'
ICON_SIGN_OUT = 'fa fa-sign-out'
ICON_SIGN_LANGUAGE = 'fa fa-sign-language'
ICON_SIGNAL = 'fa fa-signal'
ICON_SIGNING = 'fa fa-signing'
ICON_SITEMAP = 'fa fa-sitemap'
ICON_SLIDERS = 'fa fa-sliders'
ICON_SMILE_O = 'fa fa-smile-o'
ICON_SNOWFLAKE_O = 'fa fa-snowflake-o'
ICON_SOCCER_BALL_O = 'fa fa-soccer-ball-o'
ICON_SORT = 'fa fa-sort'
ICON_SORT_ALPHA_ASC = 'fa fa-sort-alpha-asc'
ICON_SORT_ALPHA_DESC = 'fa fa-sort-alpha-desc'
ICON_SORT_AMOUNT_ASC = 'fa fa-sort-amount-asc'
ICON_SORT_AMOUNT_DESC = 'fa fa-sort-amount-desc'
ICON_SORT_ASC = 'fa fa-sort-asc'
ICON_SORT_DESC = 'fa fa-sort-desc'
ICON_SORT_DOWN = 'fa fa-sort-down'
ICON_SORT_NUMERIC_ASC = 'fa fa-sort-numeric-asc'
ICON_SORT_NUMERIC_DESC = 'fa fa-sort-numeric-desc'
ICON_SORT_UP = 'fa fa-sort-up'
ICON_SPACE_SHUTTLE = 'fa fa-space-shuttle'
ICON_SPINNER = 'fa fa-spinner'
ICON_SPOON = 'fa fa-spoon'
ICON_SQUARE = 'fa fa-square'
ICON_SQUARE_O = 'fa fa-square-o'
ICON_STAR = 'fa fa-star'
ICON_STAR_HALF = 'fa fa-star-half'
ICON_STAR_HALF_EMPTY = 'fa fa-star-half-empty'
ICON_STAR_HALF_FULL = 'fa fa-star-half-full'
ICON_STAR_HALF_O = 'fa fa-star-half-o'
ICON_STAR_O = 'fa fa-star-o'
ICON_STICKY_NOTE = 'fa fa-sticky-note'
ICON_STICKY_NOTE_O = 'fa fa-sticky-note-o'
ICON_STREET_VIEW = 'fa fa-street-view'
ICON_SUITCASE = 'fa fa-suitcase'
ICON_SUN_O = 'fa fa-sun-o'
ICON_SUPPORT = 'fa fa-support'
ICON_TABLET = 'fa fa-tablet'
ICON_TACHOMETER = 'fa fa-tachometer'
ICON_TAG = 'fa fa-tag'
ICON_TAGS = 'fa fa-tags'
ICON_TASKS = 'fa fa-tasks'
ICON_TAXI = 'fa fa-taxi'
ICON_TELEVISION = 'fa fa-television'
ICON_TERMINAL = 'fa fa-terminal'
ICON_THERMOMETER = 'fa fa-thermometer'
ICON_THERMOMETER_0 = 'fa fa-thermometer-0'
ICON_THERMOMETER_1 = 'fa fa-thermometer-1'
ICON_THERMOMETER_2 = 'fa fa-thermometer-2'
ICON_THERMOMETER_3 = 'fa fa-thermometer-3'
ICON_THERMOMETER_4 = 'fa fa-thermometer-4'
ICON_THERMOMETER_EMPTY = 'fa fa-thermometer-empty'
ICON_THERMOMETER_FULL = 'fa fa-thermometer-full'
ICON_THERMOMETER_HALF = 'fa fa-thermometer-half'
ICON_THERMOMETER_QUARTER = 'fa fa-thermometer-quarter'
ICON_THERMOMETER_THREE_QUARTERS = 'fa fa-thermometer-three-quarters'
ICON_THUMB_TACK = 'fa fa-thumb-tack'
ICON_THUMBS_DOWN = 'fa fa-thumbs-down'
ICON_THUMBS_O_UP = 'fa fa-thumbs-o-up'
ICON_THUMBS_UP = 'fa fa-thumbs-up'
ICON_TICKET = 'fa fa-ticket'
ICON_TIMES = 'fa fa-times'
ICON_TIMES_CIRCLE = 'fa fa-times-circle'
ICON_TIMES_CIRCLE_O = 'fa fa-times-circle-o'
ICON_TIMES_RECTANGLE = 'fa fa-times-rectangle'
ICON_TIMES_RECTANGLE_O = 'fa fa-times-rectangle-o'
ICON_TINT = 'fa fa-tint'
ICON_TOGGLE_DOWN = 'fa fa-toggle-down'
ICON_TOGGLE_LEFT = 'fa fa-toggle-left'
ICON_TOGGLE_RIGHT = 'fa fa-toggle-right'
ICON_TOGGLE_UP = 'fa fa-toggle-up'
ICON_TOGGLE_OFF = 'fa fa-toggle-off'
ICON_TOGGLE_ON = 'fa fa-toggle-on'
ICON_TRADEMARK = 'fa fa-trademark'
ICON_TRASH = 'fa fa-trash'
ICON_TRASH_O = 'fa fa-trash-o'
ICON_TREE = 'fa fa-tree'
ICON_TROPHY = 'fa fa-trophy'
ICON_TRUCK = 'fa fa-truck'
ICON_TTY = 'fa fa-tty'
ICON_TV = 'fa fa-tv'
ICON_UMBRELLA = 'fa fa-umbrella'
ICON_UNIVERSAL_ACCESS = 'fa fa-universal-access'
ICON_UNIVERSITY = 'fa fa-university'
ICON_UNLOCK = 'fa fa-unlock'
ICON_UNLOCK_ALT = 'fa fa-unlock-alt'
ICON_UNSORTED = 'fa fa-unsorted'
ICON_UPLOAD = 'fa fa-upload'
ICON_USER = 'fa fa-user'
ICON_USER_CIRCLE = 'fa fa-user-circle'
ICON_USER_CIRCLE_O = 'fa fa-user-circle-o'
ICON_USER_O = 'fa fa-user-o'
ICON_USER_PLUS = 'fa fa-user-plus'
ICON_USER_SECRET = 'fa fa-user-secret'
ICON_USER_TIMES = 'fa fa-user-times'
ICON_USERS = 'fa fa-users'
ICON_VCARD = 'fa fa-vcard'
ICON_VCARD_O = 'fa fa-vcard-o'
ICON_VIDEO_CAMERA = 'fa fa-video-camera'
ICON_VOLUME_CONTROL_PHONE = 'fa fa-volume-control-phone'
ICON_VOLUME_DOWN = 'fa fa-volume-down'
ICON_VOLUME_OFF = 'fa fa-volume-off'
ICON_VOLUME_UP = 'fa fa-volume-up'
ICON_WARNING = 'fa fa-warning'
ICON_WHEELCHAIR = 'fa fa-wheelchair'
ICON_WHEELCHAIR_ALT = 'fa fa-wheelchair-alt'
ICON_WINDOW_CLOSE = 'fa fa-window-close'
ICON_WINDOW_CLOSE_O = 'fa fa-window-close-o'
ICON_WINDOW_MAXIMIZE = 'fa fa-window-maximize'
ICON_WINDOW_MINIMIZE = 'fa fa-window-minimize'
ICON_WINDOW_RESTORE = 'fa fa-window-restore'
ICON_WIFI = 'fa fa-wifi'
ICON_WRENCH = 'fa fa-wrench'

EFFECT_SPIN = 'fa-spin'
EFFECT_PULSE = 'fa-pulse'


if __name__ == '__main__':
  data = '''
  '''
  for rec in data.split("\n"):
    if rec.strip() != "":
      srec = rec.strip().split(" ")
      icon = srec[1].replace("fa-", "ICON_").replace("-", "_")
      print("%s = '%s'" % (icon.upper(), rec.strip()))




import re
import json
import collections
import functools
import logging

from epyk.core.js import JsUtils
from epyk.core.js import Js
from epyk.core.js.html import JsHtml
from epyk.core.js import packages
from epyk.core.js.packages import JsQuery

from epyk.core.css.styles import GrpCls
from epyk.core.html import Aria

try:  # For python 3
  import urllib.request as urllib2
  import urllib.parse as parse
except:  # For Python 2
  import urllib2
  import urllib as parse


regex = re.compile('[^a-zA-Z0-9_]')


def cleanData(value):
  """ Function to clean the javascript data to allow the use of variables """
  return regex.sub('', value.strip())


# ---------------------------------------------------------------------------------------------------------
#                                          FRAMEWORK DECORATORS
#
# ---------------------------------------------------------------------------------------------------------
def deprecated(func):
  """
  This is a decorator which can be used to mark functions
  as deprecated. It will result in a warning being emmitted
  when the function is used.
  """

  @functools.wraps(func)
  def new_func(*args, **kwargs):
    logging.warn('#########################################')
    logging.warn("Call to deprecated function {}.".format(func.__name__))
    logging.warn('#########################################')
    return func(*args, **kwargs)
  return new_func


def inprogress(func):
  @functools.wraps(func)
  def new_func(*args, **kwargs):
    # warnings.simplefilter('always', DeprecationWarning)  # turn off filter
    # warnings.warn('############################################################################')
    # warnings.warn("Call to a test function {}.".format(func.__name__), category=DeprecationWarning, stacklevel=2)
    # warnings.warn('############################################################################')
    # warnings.simplefilter('default', DeprecationWarning)  # reset filter
    return func(*args, **kwargs)

  return new_func


class Html(object):
  """
  Parent class for all the HTML components. All the function defined here are available in the children classes.
  Child class can from time to time re implement the logic but the function will always get the same meaning (namely the same signature and return)
  """
  cssCls = None
  # Those variables should not be used anymore and should be replaced by the __ ones
  # This is done in order to avoid having users to change them. Thanks to the name
  # mangling technique Python will make the change more difficult and easier to see
  reqJs, reqCss = [], []
  htmlCode, dataSrc, _code, inReport, builder_name = None, None, None, True, None

  def __init__(self, report, vals, htmlCode=None, code=None, width=None, widthUnit=None, height=None,
               heightUnit=None, globalFilter=None, dataSrc=None, options=None, profile=None, css_attrs=None):
    """ Create an python HTML object """
    self._triggerEvents, self.profile = set(), profile
    self._report, self._styleObj = report, None # The html object ID
    self._dom, self._container, self._sub_htmls, self._js, self.helper = None, None, [], None, ""
    self.jsImports = report.jsImports
    self.cssImport = report.cssImport
    self.attr = {'class': self.style.classList['main'], 'css': self.style.css.attrs}
    if css_attrs is not None:
      self.css(css_attrs)
    self.jsFncFrag, self._code, self._jsStyles, self._events = {}, code, {}, {"comp_ready": {}, 'doc_ready': {}}
    self.innerPyHTML = None
    if code is not None:
      # Control to ensure the Javascript problem due to multiple references is highlighted during the report generation
      if code in self._report.htmlRefs:
        raise Exception("Duplicated Html Code %s in the script !" % code)

      self._report.htmlRefs[code] = True
      self._report.htmlCodes[code] = self
      if code[0].isdigit() or cleanData(code) != code:
        raise Exception("htmlCode %s cannot start with a number or contain, suggestion %s " % (code, cleanData(code)))

    if htmlCode is not None:
      self._report.htmlRefs[htmlCode] = True
      if htmlCode[0].isdigit() or cleanData(htmlCode) != htmlCode:
        raise Exception("htmlCode %s cannot start with a number or contain, suggestion %s " % (htmlCode, cleanData(htmlCode)))

      self._report.htmlCodes[htmlCode] = self
      try:
        int(htmlCode[0])
        raise Exception("htmlCode cannot start with a number - %s" % htmlCode)

      except: pass

      self.htmlCode = htmlCode
      # self._report.jsGlobal.reportHtmlCode.add(htmlCode)
      if htmlCode in self._report.http:
        self.vals = self._report.http[htmlCode]

    #css = None
    self.pyStyle = None #list(getattr(self, '_%s__pyStyle' % self.__class__.__name__, []))
    if hasattr(self, '_%s__reqJs' % self.__class__.__name__):
      self.reqJs = list(getattr(self, '_%s__reqJs' % self.__class__.__name__, []))
    if hasattr(self, '_%s__reqCss' % self.__class__.__name__):
      self.reqCss = list(getattr(self, '_%s__reqCss' % self.__class__.__name__, []))
    self.pyCssCls = set()
    self.jsOnLoad, self.jsEvent, self.jsEventFnc = set(), {}, collections.defaultdict(set)
    self._vals = vals
    self.jsVal = "%s_data" % self.htmlId
    if self._report is not None:
       # Some components are not using _report because they are directly used for the display
       if self.reqJs is not None:
         for js in self.reqJs:
           self._report.jsImports.add(js)

       if self.reqCss is not None:
         for css in self.reqCss:
           self._report.cssImport.add(css)
    # Add the CSS dimension
    if width is not None:
      self.css({'width': "%s%s" % (width, widthUnit)})
    if height is not None:
      self.css('height', "%s%s" % (height, heightUnit))
    if htmlCode is not None and globalFilter is not None:
      self.filter(**globalFilter)
    if dataSrc is not None:
      self.dataSrc = dataSrc
    self.builder_name = self.builder_name if self.builder_name is not None else self.__class__.__name__

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpCls.ClassHtml(self)
    return self._styleObj

  @property
  def htmlId(self):
    if self._code is not None:
      # This is a special code used to update components but not to store the results to the breadcrumb
      # Indeed for example for components like paragraph this does not really make sense to use the htmlCode
      return self._code

    if self.htmlCode is not None:
      return self.htmlCode

    return "%s_%s" % (self.__class__.__name__.lower(), id(self))

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object
    :rtype: Js.JsBase
    """
    if self._js is None:
      self._js = Js.JsBase(self._report)
    return self._js

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtml
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtml(self, report=self._report)
    return self._dom

  def prepend_child(self, htmlObj):
    """
    Description:
    -----------
    Wrapper to the Javascript method insertChild to add an HTML component

    Usage::

      for i in range(10):
      encore = rptObj.ui.texts.label("Add Label %s" % i).css({"width": "100%", "display": 'block'})
      select.prepend_child(encore)

    Related Pages:

			https://www.w3schools.com/jsref/met_node_insertbefore.asp

    Attributes:
    ----------
    :param htmlObj: The html component
    :return: The htmlObj
    """
    self._sub_htmls.append(htmlObj)
    htmlObj.inReport = False
    # add a flag to propagate on the Javascript the fact that some child nodes will be added
    # in this case innerHYML cannot be used anymore
    self._jsStyles["_children"] = self._jsStyles.get("_children", 0) + 1
    self._report.js.addOnLoad([self.dom.insertBefore(htmlObj.dom)])
    return self

  def append_child(self, htmlObj):
    """
    Description:
    -----------
    Wrapper to the Javascript method appendChild to append an HTML component

    Usage::

      for i in range(10):
      encore = rptObj.ui.texts.label("Add Label %s" % i).css({"width": "100%", "display": 'block'})
      select.append_child(encore)

    Related Pages:

			https://www.w3schools.com/jsref/met_node_appendchild.asp

    Attributes:
    ----------
    :param htmlObj: The html component

    :return: The htmlObj
    """
    self._sub_htmls.append(htmlObj)
    htmlObj.inReport = False
    # add a flag to propagate on the Javascript the fact that some child nodes will be added
    # in this case innerHYML cannot be used anymore
    self._jsStyles["_children"] = self._jsStyles.get("_children", 0) + 1
    self._report.js.addOnLoad([self.dom.appendChild(htmlObj.dom)])
    return self

  def onReady(self, jsFncs):
    """
    Description:
    -----------
    Add set of event / actions whihc will be triggered after the build of the object.
    usually this can be used to add js functions on a chart or a table

    Usage::

      network = rptObj.ui.charts.vis.network()
    network.onReady([
      network.js.setData({"nodes": [{"id": 0, "label": "test"}], "edges": []}),
    ])

    Attributes:
    ----------
    :param jsFncs: List. Javascript function to be added once the object is built
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._report.js.addOnReady(jsFncs)

  def add_menu(self, context_menu):
    """
    Description:
    -----------
    Attach a context menu to an existing component. A context menu must have a component attached to otherwise
    the report will not be triggered

    Attributes:
    ----------
    :param context_menu: A Python context menu object
    """
    context_menu.source = self
    self._report._contextMenu[self.dom.jquery.varName] = context_menu
    return self

  def add_icon(self, text, css=None, position="before"):
    """
    Description:
    ------------
    Add an icon to the HTML object

    Usage::

      checks.title.add_icon("fas fa-align-center")

    Related Pages:
Attributes:
    ----------
    :param text: The icon reference from font awsome website
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:

    :return: The Html object
    """
    self.icon = ""
    if text is not None:
      self.icon = self._report.ui.images.icon(text).css({"margin-right": '5px'})
      if position == "before":
        self.prepend_child(self.icon)
      else:
        self.append_child(self.icon)
      #elf.icon.inReport = False
      if css is not None:
        self.icon.css(css)
    return self

  def add_label(self, text, css=None, position="before", for_=None):
    """
    Description:
    -----------
    Add an elementary label component

    Usage::

      Related Pages:

			https://www.w3schools.com/tags/tag_label.asp

    Attributes:
    ----------
    :param text: The label content
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:
    :param for_: Specifies which form element a label is bound to
    """
    self.label = ""
    if text is not None:
      self.label = self._report.ui.texts.label(text)
      if for_ is not None:
        # Attach the label to another HTML component based on the ID
        self.label.attr['for'] = for_
      if position == "before":
        self.prepend_child(self.label)
      else:
        self.append_child(self.label)
      if css == False:
        self.label.attr['css'] = {}
      elif css is not None:
        self.label.css(css)
    return self

  def add_span(self, text, css=None, position="before", i=None):
    """
    Description:
    -----------
    Add an elementary span component

    Usage::

      Related Pages:

			https://fontawesome.com/how-to-use/on-the-web/styling/layering

    Attributes:
    ----------
    :param text: The Span content
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:
    :param i:
    """
    if i is not None:
      key_attr = 'span_%s' % i
    else:
      key_attr = 'span'
    setattr(self, key_attr, '')
    if text is not None:
      setattr(self, key_attr, self._report.ui.texts.span(text))
      span = getattr(self, key_attr)
      if position == "before":
        self.prepend_child(span)
      else:
        self.append_child(span)
      if css == False:
        span.attr['css'] = {}
      elif css is not None:
        span.css(css)
    return self

  def add_link(self, text, url=None, script_name=None, report_name=None, name=None, icon=None, css=None, position="before"):
    """
    Description:
    -----------
    Add an elementary label component

    Usage::

      div = rptObj.ui.div()
    div.add_link("test.py", name="Click to go to the test report")

    Attributes:
    ----------
    :param text:
    :param url:
    :param script_name:
    :param report_name:
    :param name:
    :param icon:
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:
    """
    self.link = ""
    if url is not None or script_name is not None:
      options = {"name": name} if name is not None else {}
      if url is not None:
        self.link = self._report.ui.links.external(text, url)
      else:
        self.link = self._report.ui.links.script(text, script_name, report_name, icon=icon, options=options)
      if position == "before":
        self.prepend_child(self.link)
      else:
        self.append_child(self.link)
      if css is not None:
        self.link.css(css)
    return self

  def add_title(self, text, level=None, css=None, position="before", options=None):
    """
    Description:
    -----------
    Add an elementary title component

    Usage::

      Attributes:
    ----------
    :param text: The title content
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:
    """
    self.title = ""
    if text is not None:
      self.title = self._report.ui.texts.title(text, level=level, options=options)
      if options.get('managed', True):
        if position == "before":
          self.prepend_child(self.title)
        else:
          self.append_child(self.title)
      else:
        self.title.inReport = False
      if css == False:
        self.title.attr['css'] = {}
      elif css is not None:
        self.title.css(css)
    return self

  def add_input(self, text, css=None, attrs=None, position="before"):
    """
    Description:
    -----------
    Add an elementary input component

    Usage::

      Attributes:
    ----------
    :param text: The title content
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param attrs: Optional
    :param position:
    """
    self.input = ""
    if text is not None:
      self.input = self._report.ui.inputs.input(text)
      if position == "before":
        self.prepend_child(self.input)
      else:
        self.append_child(self.input)
      if css is not None:
        self.input.css(css)
      if attrs is not None:
        self.input.set_attrs(attrs=attrs)
    return self

  def add_checkbox(self, flag, css=None, attrs=None, position="before"):
    """
    Description:
    -----------
    Add an elementary checkbox component

    Usage::

      Attributes:
    ----------
    :param flag: Boolean. The state of the checkbox component
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param attrs: Optional
    :param position:
    """
    self.checkbox = ""
    if flag is not None:
      self.checkbox = self._report.ui.inputs.checkbox(flag)
      if position == "before":
        self.prepend_child(self.checkbox)
      else:
        self.append_child(self.checkbox)
      if css is not None:
        self.checkbox.css(css)
      if attrs is not None:
        self.checkbox.set_attrs(attrs=attrs)
    return self

  def add_helper(self, text, css=None):
    """
    Description:
    -----------
    Add an elementary helper icon

    Usage::

      Attributes:
    ----------
    :param text: The helper content
    :param css: Optional. A dictionary with the CSS style to be added to the component

    :rtype: self._report.ui.rich.info
    """
    if text is not None:
      self.helper = self._report.ui.rich.info(text)
      self.helper.inReport = False
      if css is not None:
        self.helper.css(css)
    return self

  @property
  def aria(self):
    """
    Accessible Rich Internet Applications is a [HTML] specification module.
    Web developers MAY use the ARIA role and aria-* attributes on HTML elements

    Related Pages:

			https://www.w3.org/TR/html-aria/#allowed-aria-roles-states-and-properties
    """
    return Aria.Aria(self)

  @property
  def val(self):
    """
    Description:
    -----------
    Property to get the jquery value of the HTML object in a python HTML object.
    This method can be used in any jsFunction to get the value of a component in the browser.
    This method will only be used on the javascript side, so please do not consider it in your algorithm in Python

    :returns: Javascript string with the function to get the current value of the component
    """
    return self._vals

  @property
  def content(self):
    if self.innerPyHTML is not None:
      if isinstance(self.innerPyHTML, list):
        return "".join([h.html() for h in self.innerPyHTML])

      return self.innerPyHTML.html()

    return self.val if not hasattr(self.val, "html") else self.val.html()

  def move(self):
    """
    Description:
    -----------
    Move the component to this position in the page
    """
    comp_id = id(self)
    self._report.content.remove(comp_id)
    self._report.content.append(comp_id)

  def css(self, key, value=None, reset=False):
    """
    Description:
    -----------
    Change the CSS Style of a main component. This is trying to mimic the signature of the Jquery css function

    Related Pages:
http://api.jquery.com/css/

    Attributes:
    ----------
    :param key: The key style in the CSS attributes (Can also be a dictionary)
    :param value: The value corresponding to the key style
    :param reset: Boolean

    :return: The python object itself
    """
    if reset:
      self.style.css.attrs = {}
      self.attr['css'] = self.style.css.attrs
    if value is None and isinstance(key, dict):
      # Do not add None value to the CSS otherwise it will break the page on the front end side
      css_vals = key if isinstance(key, dict) else {}
    elif value is None and not isinstance(key, dict):
      return self.attr['css'].get(key)

    else:
      if isinstance(value, tuple):
        value = value[0] if value[0] is None else "%s%s" % (value[0], value[1])
      css_vals = {key: value}
    if not 'css' in self.attr:
      self.attr['css'] = self.style.css.attrs
    for key, value in css_vals.items():
      if isinstance(value, tuple):
        if value[0] is None:
          continue

        value = "%s%s" % (value[0], value[1])
      if value is None:
        continue

      self.attr['css'][key] = value if isinstance(value, str) else json.dumps(value)
    return self

  def tooltip(self, value, location='top'):
    """
    Description:
    -----------
    Add the Tooltip feature when the mouse is over the component.
    This tooltip version is coming from Bootstrap

    Usage::

      htmlObj.tooltip("My tooltip", location="bottom")

    Related Pages:

			https://getbootstrap.com/docs/4.1/components/tooltips/

    Attributes:
    ----------
    :param value:
    :param location:

    :return: The Python object self
    """
    self.attr.update({'title': value, 'data-toggle': 'tooltip', 'data-placement': location})
    self._report._props['js']['onReady'].add("%s.tooltip()" % JsQuery.decorate_var("'[data-toggle=tooltip]'", convert_var=False))
    return self

  @packages.packageImport('bootstrap', 'bootstrap')
  def popover(self, content, title=None, options=None):
    """
    Description:
    -----------
    All the attributes will be added to the

    Related Pages:

			https://getbootstrap.com/docs/4.4/components/popovers/

    Attributes:
    ----------
    :param content: String. The tooltip content
    :param title: String. The tooltip title
    :param options: Dictionary all the options to be attached to the component
    """
    self.attr["data-content"] = content
    if title is not None:
      self.attr["data-title"] = title
    if options is not None:
      for k, v in options.items():
        self.attr["data-%s" % k] = title
    self.attr["data-toggle"] = 'popover'
    self._report._props['js']['onReady'].add("%s.popover()" % JsQuery.decorate_var("'[data-toggle=popover]'", convert_var=False))
    return self

  def add_options(self, options=None, name=None, value=None):
    """
    Description:
    -----------
    Change the Javascript options of the component.
    This will change the options sent to the Javascript

    Attributes:
    ----------
    :param options: Dictionary with the options
    :param name: String. The key
    :param value: String. The value

    :return: self to allow the chains
    """
    if options is None and name is None:
      raise Exception("Either the attrs or the name should be specified")

    if options is None:
      options = {name: value}
    for k, v in options.items():
      self._jsStyles[k] = v
    return self

  def set_attrs(self, attrs=None, name=None, value=None):
    """
    Description:
    -----------
    Function to update the internal dictionary of object attributes. Those attributes will be used when the HTML component will be defined.
    Basically all sort of attributes can be defined here: CSS attributes, but also data, name...
    Either the attrs or the tuple (name, value) can be used to add information to the dom object.

    All the attributes should be Python object which are ready to use on the Javascript side.
    For example True should be written 'true'

    Tips: It is possible to use the data- attributes to store any kind of information in the dom.

    Related Pages:
Attributes:
    ----------
    :param attrs: A python dictionary with the attributes
    :param name: A python string with the name of the attribute
    :param value: A python string with the value of the attribute
    """
    if attrs is None and name is None:
      raise Exception("Either the attrs or the name should be specified")

    if attrs is None:
      attrs = {name: value}
    for k, v in attrs.items():
      if k == 'css':
        # Section for the Style attributes
        if v is None:
          self.style.clear_style()
          continue

        if not 'css' in self.attr:
          self.attr['css'] = dict(v)
        else:
          self.attr['css'].update(v)
      elif k == 'class':
        self.style.clear()
        if v is None:
          continue

        if not isinstance(v, set):
          v = set(v.split(" "))
        for c in v:
          self.attr['class'].add(c)
      else:
        # Section for all the other attributes#
        if v is not None:
          self.attr[k] = v
    return self

  def get_attrs(self, withId=True, pyClassNames=None):
    """
    Description:
    -----------
    Return the string line with all the attributes

    All the attributes in the div should use double quote and not simple quote to be consistent everywhere in the framework
    and also in the javascript. If there is an inconsistency, the aggregation of the string fragments will not work

    Attributes:
    ----------
    :param withId:
    :param pyClassNames:

    :return: A string with the dom attributes
    """
    cssStyle, cssClass, classData = '', '', ''
    if 'css' in self.attr:
      styles = ";".join(["%s:%s" % (key, val) for key, val in self.attr["css"].items()])
      if styles:
        cssStyle = 'style="%s"' % styles
    if 'class' in self.attr and len(self.attr['class']) > 0 and classData:
      if pyClassNames is not None:
        # Need to merge in the class attribute some static classes coming from external CSS Styles sheets
        # and the static python classes defined on demand in the header of your report
        # self._report.cssObj.getClsTag(pyClassNames)[:-1] to remove the ' generated in the module automatically
        cssClass = self._report.style.getClsTag(pyClassNames.clsMap).replace('class="', 'class="%s ')
        if cssClass:
          cssClass %= classData
        else:
          cssClass = 'class="%s"' % classData
      else:
        cssClass = 'class="%s"' % classData
    elif pyClassNames is not None:
      pyClsNames = [cls.get_ref() if hasattr(cls, 'get_ref') else cls for cls in pyClassNames['main']]
      cssClass = 'class="%s"' % " ".join(pyClsNames) if len(pyClsNames) > 0 else ""

    if withId:
      self.attr['id'] = self.htmlId
    html_tags = ['%s="%s"' % (key, str(val).replace('"', "'")) if val is not None else key for key, val in self.attr.items() if key not in ('css', 'class')]
    for tag in [cssStyle, cssClass]:
      if tag:
        html_tags.append(tag)
    str_tag = " ".join(html_tags)
    return str_tag.strip()

  # -------------------------------------------------------------
  # Javascript Event wrappers
  #
  def on(self, event, jsFncs, profile=False):
    """
    Description:
    -----------
    Add an event to the document ready function.
    This is to mimic the Jquery on function.

    Related Pages:

			https://www.w3schools.com/jquery/event_on.asp
    https://www.w3schools.com/js/js_htmldom_eventlistener.asp
    https://www.w3schools.com/jsref/dom_obj_event.asp

    Attributes:
    ----------
    :param event: A string with the Javascript event type from the dom_obj_event.asp
    :param jsFncs: A Javascript Python function
    :param profile: A Boolean. Set to true to get the profile for the function on the Javascript console

    :return: self to allow the chains
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    # JsUtils.jsConvertFncs needs to be applied in order to freeze the function
    # span.on("mouseover", span.dom.css("color", "red"))
    # span.on("mouseleave", span.dom.css("color", "blue"))
    self._events['doc_ready'].setdefault(event, {}).setdefault("content", []).extend(JsUtils.jsConvertFncs(jsFncs))
    self._events['doc_ready'][event]['profile'] = profile
    return self

  def drop(self, jsFncs, preventDefault=True, profile=False):
    """
    Description:
    -----------
    Add a drag and drop property to the element

    Usage::

      d = rptObj.ui.div()
    d.drop([rptObj.js.objects.data.toRecord([1, 2, 3, 4], "result")])

    Attributes:
    ----------
    :param jsFncs:
    :param preventDefault: Boolean.
    :param profile:

    :return: Return self to allow the chaining
    """
    dft_fnc = ""
    if preventDefault:
      dft_fnc = self.js.objects.event.preventDefault()
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    str_fncs = JsUtils.jsConvertFncs(["var data = %s" % self.js.objects.event.dataTransfer.text] + jsFncs, toStr=True)
    self.attr["ondrop"] = "(function(event){%s; %s; return false})(event)" % (dft_fnc, str_fncs)
    self.attr["ondragover"] = "(function(event){%s})(event)" % dft_fnc
    return self

  def hover(self, jsFncs, profile=False):
    return self.on("mouseover", jsFncs, profile)

  def click(self, jsFncs, profile=False):
    return self.on("click", jsFncs, profile)

  def mouse(self, on_fncs=None, out_fncs=None, profile=False):
    """
    Description:
    -----------
    Wrapper function fot the mouse event.
    This function will cover the on mouse hover event and mouse out.

    More specific events are possible using the generic out function

    Usage::

      span.mouse([
        span.dom.css("color", "red"),
        span.dom.css("cursor", "pointer").r],
        span.dom.css("color", "blue").r)

    Related Pages:
Attributes:
    ----------
    :param on_fncs: Array or String of Javascript events
    :param out_fncs: Array or String of Javascript events

    :return: self to allow the chains
    """
    self.style.css.cursor = 'pointer'
    if on_fncs is not None:
      self.on("mouseenter", on_fncs, profile)
    if out_fncs is not None:
      self.on("mouseleave", out_fncs, profile)
    return self

  def contextMenu(self, menu, jsFncs, profile=False):
    """
    Description:
    -----------
    Attach a context menu to a component and set a function to called before the display

Attributes:
    ----------
    :param menu:
    :param jsFncs:
    :param profile:
    """
    menu.source = self
    # event.stopPropagation(); %(jqId)s.css({left: event.pageX + 1, top: event.pageY + 1, display: 'block'}); event.preventDefault()
    new_js_fncs = jsFncs + [self._report.js.objects.mouseEvent.stopPropagation(),
                   menu.dom.css({"display": 'block', 'left': self._report.js.objects.mouseEvent.clientX + "'px'",
                                 'top': self._report.js.objects.mouseEvent.clientY + "'px'"}),
                   self._report.js.objects.mouseEvent.preventDefault()]
    self.on("contextmenu", new_js_fncs, profile)
    return self

  # -------------------------------------------------------------
  # Builder functions
  #
  @property
  def _js__builder__(self):
    raise Exception("Constructor must be defined in %s" % self.__class__.__name__)

  def build(self, data=None, options=None, profile=False):
    if not self.builder_name:
      raise Exception("No builder defined for this HTML component %s" % self.__class__.__name__)

    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[self.builder_name] = "function %s(htmlObj, data, options){%s}" % (
    self.builder_name, self._js__builder__)

    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    options, js_options = options or self._jsStyles, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s(%s, %s, %s)" % (self.builder_name, self.dom.varId, js_data, "{%s}" % ",".join(js_options))

  def refresh(self):
    # self._report._props.setdefault('js', {}).setdefault("builders", []).append(refresh_js)
    return self.build(self.val, self._jsStyles)

  def onDocumentLoadContextmenu(self):
    self._report.jsGlobal.fnc("ContextMenu(htmlObj, data, markdownFnc)",
        '''
        $('#popup').empty(); $('#popup').append('<ul style="width:100%%;height:100%%;margin:0;padding:0"></ul>');
        var listMenu = $('#popup').find('ul');
        data.forEach(function(rec){
          if ('title' in rec) {
            listMenu.append('<li class="list-group-item" style="cursor:cursor;width:100%%;display:inline-block;padding:5px 5px 2px 10px;font-weight:bold;color:white;background:%(color)s">' + rec.title + '</li> ');
          } else {
            if (rec.url != undefined) { var content = '<a href="' + rec.url + '" style="color:black">' + rec.label + '</a>' ;} else {var content = rec.label;};
            listMenu.append('<li class="list-group-item" style="cursor:pointer;width:100%%;display:inline-block;padding:2px 5px 2px 10px">' + content + '</li> '); }
        });
        if (markdownFnc != false) {
          listMenu.append('<li class="list-group-item" style="cursor:cursor;width:100%%;display:inline-block;padding:5px 5px 2px 10px;font-weight:bold;color:white;background:%(color)s">MarkDown</li> ');
          listMenu.append('<li onclick="CopyMarkDown(\\''+ markdownFnc +'\\');" class="list-group-item" style="cursor:pointer;width:100%%;display:inline-block;padding:2px 5px 2px 10px"><i class="fas fa-thumbtack"></i>&nbsp;&nbsp;Copy MarkDown</li> ');};
        $('#popup').css({'padding': '0', 'width': '200px'});
        $('#popup').show()''' % {'color': self._report.theme.colors[9]})

  def paste(self, jsFnc):
    """ Generic click function """
    self._report.jsOnLoadFnc.add('''%(jqId)s.on('paste', function(event) { 
       var data;
       if (window.clipboardData && window.clipboardData.getData) { // IE
            data = window.clipboardData.getData('Text'); }
        else if (event.originalEvent.clipboardData && event.originalEvent.clipboardData.getData) { // other browsers
            data = event.originalEvent.clipboardData.getData('text/plain')} 
        %(jsFnc)s 
      })''' % {'jqId': self.jqId, 'jsFnc': jsFnc})

  def filter(self, jsId, colName, allSelected=True, filterGrp=None, operation="=", itemType="string"):
    filterObj = {"operation": operation, 'itemType': itemType, 'allIfEmpty': allSelected, 'colName': colName, 'val': self.val, 'typeVal': 'js'}
    self._report.jsSources.setdefault(jsId, {}).setdefault('_filters', {})[self.htmlCode] = filterObj
    return self

  # -------------------------------------------------------------------------------------------------------------------
  #                    OUTPUT METHODS
  # -------------------------------------------------------------------------------------------------------------------
  def __str__(self):
    """
    Description:
    -----------
    Apply the corresponding function to build the HTML result.
    This function is very specific and it has to be defined in each class.
    """
    raise NotImplementedError('subclasses must override __str__()!')

  def to_word(self, document):
    """
    Description:
    -----------
    Apply the corresponding function to produce the same result in a word document.
    This function is very specific and it has to be defined in each class.

    Related Pages:
http://python-docx.readthedocs.io/en/latest/
    """
    raise NotImplementedError('''
      subclasses must override to_word(), %s !
      Go to http://python-docx.readthedocs.io/en/latest/user/quickstart.html for more details  
    ''' % self.__class__.__name__)

  def to_xls(self, workbook, worksheet, cursor):
    """
    Description:
    -----------
    Apply the corresponding function to produce the same result in a word document.
    This function is very specific and it has to be defined in each class.

    Related Pages:

			https://xlsxwriter.readthedocs.io/
    """
    raise NotImplementedError('''
      subclasses must override to_xls(), %s !
      Go to https://xlsxwriter.readthedocs.io/working_with_tables.html for more details  
    ''' % self.__class__.__name__)

  def html(self):
    str_result = []
    for htmlObj in self._sub_htmls:
      str_result.append(htmlObj.html())
    if self.helper != "":
      self.helper.html()
    #if self.builder_name:
    #  self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    for c in self.style.get_classes()['main']:
      if hasattr(c, 'get_ref'):
        self.pyCssCls.add(c.get_ref())
    str_result.append(str(self))
    return "".join(str_result)


class Body(Html):
  name = "Body"

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpCls.ClassPage(self)
    return self._styleObj

  @property
  def htmlId(self):
    return "body"

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtml
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtml(self, report=self._report)
      self._dom.varName = "document.body"
    return self._dom

  def onReady(self, jsFncs):
    """
    Description:
    -----------
    Add set of event / actions whihc will be triggered after the build of the object.
    usually this can be used to add js functions on a chart or a table

    Usage::

      network = rptObj.ui.charts.vis.network()
    network.onReady([
      network.js.setData({"nodes": [{"id": 0, "label": "test"}], "edges": []}),
    ])

    Attributes:
    ----------
    :param jsFncs: List. Javascript function to be added once the object is built
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._report.js.addOnReady(jsFncs)

  def onLoad(self, jsFncs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def set_content(self, report, page_content):
    """
    Description:
    ------------
    Function to allow the templating of the report.
    This can be overridden by a generic class which can be shared within a set of report

    Attributes:
    ----------
    :param report: Report. The main report object
    :param page_content: String. The html content of the page

    :return: The Body HTML object
    """
    self._html_content = page_content
    return self

  def set_background(self, start_color=None, end_color=None, from_theme=False):
    """
    Description:
    ------------
    Change the body background color

    Usage::

      rptObj.body.set_background("#101626", "#374F67")

    Attributes:
    ----------
    :param start_color:
    :param end_color:
    """
    if from_theme or (start_color is None and end_color is None):
      self.style.css.background = "linear-gradient(%s 0%%, %s 100%%)" % (self._report.theme.colors[-1], self._report.theme.colors[2])
    elif end_color is not None:
      self.style.css.background = "linear-gradient(%s 0%%, %s 100%%)" % (start_color, end_color)
    else:
      self.style.css.background = start_color
    self.style.css.background_repeat = "no-repeat"
    self.style.css.background_color = self._report.theme.colors[2]

  def __str__(self):
    return '<body %s>%s</body>' % (self.get_attrs(pyClassNames=self.style.get_classes(), withId=False), self._html_content)




from epyk.core.js.html import JsHtml
from epyk.core.js import JsUtils

from epyk.core.js.primitives import JsObjects
from epyk.core.js.objects.JsNodeDom import JsDoms


class Tags(JsHtml.JsHtmlRich):

  @property
  def content(self):
    """
    Description:
    -----------
    Returns the list of data available on the filters panel
    """
    return JsHtml.ContentFormatters(self._report, '''
      (function(dom){var content = []; dom.childNodes.forEach(function(rec){content.push(rec.textContent)}); return content})(%s)
      ''' % self.querySelector("div[name=panel]"))

  def is_duplicated(self, text):
    """
    Description:
    ------------
    Check the duplicates in the filter panel
    """
    text = JsUtils.jsConvertData(text, None)
    return JsObjects.JsObjects.get(''' 
      (function(dom){var index = -1; var children = dom.childNodes; var count = 0;
        for(child in children){if(children[child].textContent == %s){
            index = count; break; }; count++; }; return index})(%s)''' % (text, self.querySelector("div[name=panel]")))

  def hide(self):
    """
    Description:
    ------------
    Hide the filters panel
    """
    return self.querySelector("div[name=panel]").show()

  def show(self):
    """
    Description:
    ------------
    Show the filters panel
    """
    return self.querySelector("div[name=panel]").show()

  def toggle(self):
    """
    Description:
    ------------
    Toggle the display of the filters panel
    """
    return self.querySelector("div[name=panel]").toggle()

  def add(self, text, no_duplicte=True):
    """
    Description:
    ------------
    Add item on the filters panel

    Attributes:
    ----------
    :param text: String. The value to be added on the filter panel
    """
    text = JsUtils.jsConvertData(text, None)
    if no_duplicte:
      return JsObjects.JsObjects.get(''' 
      if (%(duplicated)s == -1){
      var div = document.createElement("div"); div.innerHTML = %(text)s; div.style.width = 'auto'; div.style.display = 'inline'; %(css)s;
      var icon = document.createElement("i"); icon.classList.add('fas'); icon.classList.add('fa-times'); 
      icon.addEventListener('click', function(event) {div.remove()}) ;div.appendChild(icon); %(icon_css)s;
      %(panel)s.appendChild(div); }
      ''' % {'duplicated': self.is_duplicated(text),
        'panel': self.querySelector("div[name=panel]"), 'text': text, 'css': JsDoms.get(varName="div").css(self._src.options.item_css).r,
        'icon_css': JsDoms.get(varName="icon").css({"vertical-align": "middle", "display": "inline-block",
                              "margin": "auto 0", "padding": "auto 0", "margin-left": "2px", "cursor": "pointer"}).r})

    return JsObjects.JsObjects.get(''' 
      var div = document.createElement("div"); div.innerHTML = %(text)s; div.style.width = 'auto'; div.style.display = 'inline'; %(css)s;
      var icon = document.createElement("i"); icon.classList.add('fas'); icon.classList.add('fa-times'); 
      icon.addEventListener('click', function(event) {div.remove()}) ;div.appendChild(icon); %(icon_css)s;
      %(panel)s.appendChild(div);
      ''' % {'panel': self.querySelector("div[name=panel]"), 'text': text, 'css': JsDoms.get(varName="div").css(self._src.options.item_css).r,
             'icon_css': JsDoms.get(varName="icon").css({"vertical-align": "middle", "display": "inline-block",
                              "margin": "auto 0", "padding": "auto 0", "margin-left": "2px", "cursor": "pointer"}).r})

  def clear(self):
    """
    Description:
    ------------
    Clear the content of the fitlers panel
    """
    return self.querySelector("div[name=panel]").empty()

  def remove(self, text):
    """
    Description:
    ------------
    Remove an item from the filters panel

    Attributes:
    ----------
    :param name: String. The test of the items to be removed
    """
    return JsObjects.JsObjects.get('''var itemPos = %(duplicated)s; if (itemPos >= 0){ %(panel)s.childNodes[itemPos].remove()}
      ''' % {'duplicated': self.is_duplicated(text), 'panel': self.querySelector("div[name=panel]")})




from epyk.core.js import JsUtils

from epyk.core.js.packages import JsPackage
from epyk.core.js.packages import DataAttrs


class JsNvd3Axis(object):

  def __init__(self, id):
    self._selector = id
    self._js = []

  def axisLabel(self, text):
    """
    Chart axis settings

    Example
    chart.xAxis.axisLabel('Time (ms)')

    :param text:
    """
    text = JsUtils.jsConvertData(text, None)
    self._js.append("axisLabel(%s)" % text)
    return self

  def tickFormat(self, jsFnc):
    """
    Chart axis settings

    Example
    chart.xAxis.tickFormat(d3.format(',r'))

    :param jsFnc:
    """
    jsFnc = JsUtils.jsConvertData(jsFnc, None)
    self._js.append("tickFormat(%s)" % jsFnc)
    return self

  def tickNumberFormat(self):
    self._js.append("tickFormat(function(d,i){ return d3.format(',.1f')(d); })")
    return self

  def tickCurrencyFormat(self, currency):
    self._js.append("tickFormat(function(d,i){ return '%s' + d3.format(',.1f')(d); })" % currency)
    return self

  def tickDateFormat(self):
    self._js.append("tickFormat(function(d,i){ return d3.time.format('%x')(new Date(new Date() - (20000 * 86400000) + (d * 86400000))); })")
    return self

  def tickValues(self, values):
    """
    Chart axis settings

    Example
    chart.xAxis.tickValues(10)

    :param values:
    """
    values = JsUtils.jsConvertData(values, None)
    self._js.append("tickValues(%s)" % values)
    return self

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    if len(self._js) == 0:
      return self._selector

    strData = "%(jqId)s.%(items)s" % {'jqId': self._selector, 'items': ".".join(self._js)}
    self._js = [] # empty the stack
    return strData


class JsNvd3Utils(object):
  """

  """

  def windowResize(self, jsFnc):
    """
    Update the chart when window resizes.

    :param jsFnc:

    :return:
    """
    return "nv.utils.windowResize(%s)" % jsFnc


class JsNvd3(JsPackage):
  lib_alias = {'js': 'nvd3', 'css': 'nvd3'}

  class __internal(object):
    # By default it will attach eveything to the body
    jqId, jsImports, cssImport = 'd3.select("body")', set([]), set([])

  def __init__(self, src=None, varName=None, setVar=True):
    self.src = src if src is not None else self.__internal()
    self._selector = "nv.models.%s()" % self.chartFnc
    self.varName, self.setVar = varName, setVar
    self.src.jsImports.add(self.lib_alias['js'])
    self.src.cssImport.add(self.lib_alias['css'])
    self._js, self._xaxis, self._yaxis, self._u = [[]], None, None, {}
    self._js_enums = {}

  def version(self, ver):
    """
    Change the package version number

    Example
    bar.chart.version("1.11.0")

    :param ver: String. The package versions example 1.11.0
    """
    self.src._props.setdefault("packages", {})[self.lib_alias] = ver
    return self

  def set_var(self, flag):
    self.setVar = flag
    return self

  @property
  def varId(self):
    """
    The Javascript and Python reference ID

    :return: The Javascript String of the object variable name
    """
    return self._selector if self.varName is None else self.varName

  def options(self, opts):
    pass

  def width(self, value):
    """
    The width the graph or component created inside the SVG should be made.
    The width of the container element (normally the svg itself)

    Related Pages:

			https://nvd3-community.github.io/nvd3/examples/documentation.html

    :param value: A python integer

    :return:
    """
    return self.fnc("width(%s)" % value)

  def height(self, value):
    """
    The height the graph or component created inside the SVG should be made.
    The height of the container element (normally the svg itself)

    Related Pages:

			https://nvd3-community.github.io/nvd3/examples/documentation.html

    :param value:

    :return:
    """
    return self.fnc("height(%s)" % value)

  def margin(self, options):
    """
    Object containing the margins for the chart or component.
    You can specify only certain margins in the object to change just those parts.
    Default options: { "top": 15, "right": 10, "bottom": 50, "left": 60 }

    Related Pages:

			https://nvd3-community.github.io/nvd3/examples/documentation.html

    :param options: A python dictionary with the options
    """
    options = JsUtils.jsConvertData(options, None)
    self._js.append("margin(%s)" % options)
    return self

  def useInteractiveGuideline(self, flag):
    """
    Tooltips which show all data points.

    :param flag:
    :return:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("useInteractiveGuideline(%s)" % flag)

  def transitionDuration(self, time):
    """

    :param time:
    """
    time = JsUtils.jsConvertData(time, None)
    return self.fnc("transitionDuration(%s)" % time)

  def showLegend(self, flag):
    """

    :param flag:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("showLegend(%s)" % flag)

  @property
  def xAxis(self):
    """

    :return:
    """
    if self._xaxis is None:
      self._xaxis = JsNvd3Axis("%s.xAxis" % self.varName)
    return self._xaxis

  @property
  def yAxis(self):
    """

    :return:
    """
    if self._yaxis is None:
      self._yaxis = JsNvd3Axis("%s.yAxis" % self.varName)
    return self._yaxis

  def showYAxis(self, flag):
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("showYAxis(%s)" % flag)

  def showXAxis(self, flag):
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("showXAxis(%s)" % flag)

  def update(self):
    pass

  def showControls(self, flag):
    """
    Allow user to choose 'Stacked', 'Stream', 'Expanded' mode.

    :param flag:
    :return:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("showControls(%s)" % flag)

  def noData(self):
    pass


class JsNvd3Area(JsNvd3):
  chartFnc = "stackedAreaChart"

  def x(self, column=None, jsFnc=None):
    if column is not None:
      return self.fnc("x(function(d){return d.%s})" % column)

    elif jsFnc is not None:
      jsFnc = JsUtils.jsConvertFncs(jsFnc)
      return self.fnc("x(%s)" % jsFnc)

    return self

  def y(self, column=None, jsFnc=None):
    if column is not None:
      return self.fnc("y(function(d){return d.%s})" % column)

    elif jsFnc is not None:
      jsFnc = JsUtils.jsConvertFncs(jsFnc)
      return self.fnc("y(%s)" % jsFnc)

    return self

  def rotateLabels(self, value):
    pass

  def reduceXTicks(self, flag):
    pass

  def rightAlignYAxis(self, flag):
    """
    Move the y-axis to the right side

    :param flag:
    :return:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("rightAlignYAxis(%s)" % flag)

  def clipEdge(self, flag):
    """

    :param flag:
    :return:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc("clipEdge(%s)" % flag)

  def controlLabels(self, attrs):
    """

    controlLabels({"stacked": "Stacked"})

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/stackedAreaChart.html

    :param attrs:
    """
    self.fnc("controlLabels(%s)" % JsUtils.jsConvertData(attrs, None))
    return self


class JsNvd3ParallelCoordinates(JsNvd3):
  chartFnc = "parallelCoordinates"

  def dimensionNames(self, categories):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/parallelCoordinates.html

    :param categories:
    """
    self.fnc("dimensionNames(%s)" % JsUtils.jsConvertData(categories, None))
    return self

  def dimensionFormats(self, formats):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/parallelCoordinates.html

    :param formats:
    """
    self.fnc("dimensionFormats(%s)" % JsUtils.jsConvertData(formats, None))
    return self

  def lineTension(self, val):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/parallelCoordinates.html

    :param val:
    """
    self.fnc("lineTension(%s)" % JsUtils.jsConvertData(val, None))
    return self


class JsNvd3CandlestickBar(JsNvd3):
  chartFnc = "candlestickBarChart"

  def x(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("x(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("x(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def y(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("y(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("y(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self


class JsNvd3OhlcBar(JsNvd3):
  chartFnc = "ohlcBarChart"

  def x(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("x(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("x(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def y(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("y(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("y(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self


class JsNvd3Sunburst(JsNvd3):
  chartFnc = "sunburstChart"


class JsNvd3BoxPlot(JsNvd3):
  chartFnc = "boxPlotChart"

  def x(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("x(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("x(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def staggerLabels(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlot.html

    :param flag:
    """
    self.fnc("staggerLabels(%s)" % JsUtils.jsConvertData(flag, None))
    return self

  def yDomain(self, range):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlot.html

    :param range:
    """
    self.fnc("yDomain(%s)" % JsUtils.jsConvertData(range, None))
    return self

  def maxBoxWidth(self, value):
    """
    Prevent boxes from being incredibly wide

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param value:
    """
    self.fnc("maxBoxWidth(%s)" % JsUtils.jsConvertData(value, None))
    return self

  def itemColor(self, seriesColor):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param seriesColor:
    """
    self.fnc("itemColor(function (d) { return d[%s] })" % JsUtils.jsConvertData(seriesColor, None))
    return self

  def outliers(self, outlData):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param outlData:
    """
    self.fnc("outliers(function (d) { return d[%s] })" % JsUtils.jsConvertData(outlData, None))
    return self

  def outlierValue(self, data):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param data:
    """
    self.fnc("outlierValue(function (d) { return d[%s] })" % JsUtils.jsConvertData(data, None))
    return self

  def outlierColor(self, color):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param color:
    """
    self.fnc("outlierColor(function (d) { return d[%s] })" % JsUtils.jsConvertData(color, None))
    return self

  def q1(self, q1_col):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param q1_col:
    """
    self.fnc("q1(function (d) { return d[%s] })" % JsUtils.jsConvertData(q1_col, None))
    return self

  def q2(self, q2_col):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param q2_col:
    """
    self.fnc("q2(function (d) { return d[%s] })" % JsUtils.jsConvertData(q2_col, None))
    return self

  def q3(self, q3_col):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param q3_col:
    """
    self.fnc("q3(function (d) { return d[%s] })" % JsUtils.jsConvertData(q3_col, None))
    return self

  def wl(self, wl_col):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param wl_col:
    """
    self.fnc("wl(function (d) { return d[%s] })" % JsUtils.jsConvertData(wl_col, None))
    return self

  def wh(self, wh_col):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param wh_col:
    """
    self.fnc("wh(function (d) { return d[%s] })" % JsUtils.jsConvertData(wh_col, None))
    return self

  def outlierLabel(self):
    pass


class JsNvd3Bar(JsNvd3):
  chartFnc = "discreteBarChart"

  def x(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("x(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("x(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def y(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("y(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("y(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def color(self, colors):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/TimeSeries.html

    :param colors:
    """
    self.fnc("color(%s)" % JsUtils.jsConvertData(colors, None))
    return self

  def rotateLabels(self, value):
    """
    Rotates the X axis labels by the specified degree.

    Related Pages:

			https://nvd3-community.github.io/nvd3/examples/documentation.html

    :param value:
    """
    self.fnc("rotateLabels(%s)" % value)
    return self

  def reduceXTicks(self, flag):
    pass

  def staggerLabels(self, flag):
    pass

  def tooltips(self, flag):
    pass

  def showValues(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/discreteBarChart.html

    :param flag:
    """
    self.fnc("showValues(%s)" % JsUtils.jsConvertData(range, None))
    return self

  def groupSpacing(self, value):
    pass


class JsNvd3MultiBar(JsNvd3):
  chartFnc = "multiBarChart"

  def barColor(self, colors):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/multiBarChart.html

    :param colors:
    """
    self.fnc("barColor(%s)" % JsUtils.jsConvertData(colors, None))
    return self

  def stacked(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/multiBarChart2.html

    :param flag:
    """
    self.fnc("stacked(%s)" % JsUtils.jsConvertData(range, None))
    return self


class JsNvd3MultiBarHorizontal(JsNvd3Bar):
  chartFnc = "multiBarHorizontalChart"

  def yErr(self):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/multiBarHorizontalChart.html

    """
    pass

  def barColor(self, colors):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/multiBarHorizontalChart.html

    :param colors:
    """
    self.fnc("barColor(%s)" % JsUtils.jsConvertData(colors, None))
    return self


class JsNvd3Multi(JsNvd3Bar):
  chartFnc = "multiChart"


class JsNvd3Line(JsNvd3Bar):
  chartFnc = "lineChart"


class JsNvd3Scatter(JsNvd3Bar):
  chartFnc = "scatterChart"

  def showDistX(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/scatterChart.html

    :param flag:
    """
    self.fnc("showDistX(%s)" % JsUtils.jsConvertData(flag, None))
    return self

  def showDistY(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/scatterChart.html

    :param flag:
    """
    self.fnc("showDistY(%s)" % JsUtils.jsConvertData(flag, None))
    return self

  def useVoronoi(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/scatterChart.html

    :param flag:
    """
    self.fnc("useVoronoi(%s)" % JsUtils.jsConvertData(flag, None))
    return self


class JsNvd3LineWithFocus(JsNvd3Line):
  chartFnc = "lineWithFocusChart"

  def brushExtent(self, range):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/lineWithFocusChart.html

    :param range:
    """
    self.fnc("brushExtent(%s)" % JsUtils.jsConvertData(range, None))
    return self


class JsNvd3CumulativeLine(JsNvd3Line):
  chartFnc = "cumulativeLineChart"

  def average(self, mean):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/cumulativeLineChart.html

    :param mean: String. The column name corresponding to the mean value
    """
    self.fnc("average(function(d) { return d.[%s] / 100; })" % JsUtils.jsConvertData(mean, None))
    return self

  def clipVoronoi(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/cumulativeLineChart.html

    :param flag:
    """
    self.fnc("clipVoronoi(%s)" % JsUtils.jsConvertData(flag, None))
    return self


class JsNvd3LinePlusBar(JsNvd3Bar):
  chartFnc = "linePlusBarChart"

  def legendRightAxisHint(self, text):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/linePlusBarChart.html

    :param text:
    """
    self.fnc("legendRightAxisHint(%s)" % JsUtils.jsConvertData(text, None))
    return self

  def forceY(self, indices):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/linePlusBarChart.html

    :param indices:
    """
    self.fnc("forceY(%s)" % JsUtils.jsConvertData(indices, None))
    return self

  def padData(self, flag):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/linePlusBarChart.html

    :param flag:
    """
    self.fnc("padData(%s)" % JsUtils.jsConvertData(flag, None))
    return self


class JsNvd3HistoricalBar(JsNvd3Bar):
  chartFnc = "historicalBarChart"

  def xScale(self, d3fnc):
    """
    use a time scale instead of plain numbers in order to get nice round default values in the axis

    xScale(d3.time.scale())

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/TimeSeries.html

    :param d3fnc:
    """
    self.fnc("xScale(%s)" % JsUtils.jsConvertData(d3fnc, None))
    return self

  def forceX(self, values):
    """
    fix half-bar problem on the first and last bars

    :param values:
    :return:
    """
    self.fnc("forceX(%s)" % JsUtils.jsConvertData(values, None))
    return self


class JsDataArcRadius(DataAttrs):

  @property
  def inner(self):
    """
    """
    return self._attrs["inner"]

  @inner.setter
  def inner(self, time):
    self._attrs["inner"] = time

  @property
  def outer(self):
    """
    """
    return self._attrs["outer"]

  @outer.setter
  def outer(self, time):
    self._attrs["outer"] = time


class JsNvd3Pie(JsNvd3):
  chartFnc = "pieChart"

  def arcsRadius(self, values=None):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/monitoringChart.html
    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/monitoringChart.html

    :param values:
    """
    if values is not None:
      for a in values:
        arc = self.fnc_enum('arcsRadius', JsDataArcRadius)
        for k, v in a.items():
          setattr(arc, k, v)
      return self

    return self.fnc_enum('arcsRadius', JsDataArcRadius)

  def x(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("x(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("x(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def y(self, column=None, jsFnc=None):
    if column is not None:
      self.fnc("y(function(d){return d.%s})" % column)
    elif jsFnc is not None:
      self.fnc("y(%s)" % JsUtils.jsConvertFncs(jsFnc))
    return self

  def showLabels(self, flag):
    """
    Description:
    ------------
    Display pie labels

    Related Pages:
http://nvd3.org/examples/pie.html

    Attributes:
    ----------
    :param flag:
    """
    self.fnc("showLabels(%s)" % JsUtils.jsConvertData(flag, None))
    return self

  def labelThreshold(self, value):
    """
    Description:
    ------------
    Configure the minimum slice size for labels to show up

    Related Pages:
http://nvd3.org/examples/pie.html

    Attributes:
    ----------
    :param value:
    """
    self.fnc("labelThreshold(%s)" % JsUtils.jsConvertData(value, None))
    return self

  def labelSunbeamLayout(self, flag):
    """
    Change the label orientation for each category

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/monitoringChart.html

    :param flag:
    """
    self.fnc("labelSunbeamLayout(%s)" % JsUtils.jsConvertData(flag, None))
    return self



  def labelType(self, text):
    pass

  def donutLabelsOutside(self, flag):
    """

    :param flag:
    """
    self.fnc("donutLabelsOutside(%s)" % JsUtils.jsConvertData(flag, None))
    return self



  def color(self, colors):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/TimeSeries.html

    :param colors:
    """
    self.fnc("color(%s)" % JsUtils.jsConvertData(colors, None))
    return self

  def donut(self, flag):
    """
    Description:
    ------------
    Turn on Donut mode. Makes pie chart look tasty!

    Related Pages:
http://nvd3.org/examples/pie.html

    Attributes:
    ----------
    :param flag:
    """
    value = "donut(%s)" % JsUtils.jsConvertData(flag, None)
    if not value in self._js[-1]:
      self.fnc(value)
    return self

  def startAngle(self, angle):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/donutChart.html
    """
    self.fnc("startAngle(function(d) { return d.startAngle/2 -Math.PI/2; } )")
    return self

  def endAngle(self, angle):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/donutChart.html

    """
    self.fnc("endAngle(function(d) { return d.startAngle/2 -Math.PI/2 })")
    return self



  def half(self):
    pass

  def donutRatio(self, value):
    """
    Description:
    ------------
    Configure how big you want the donut hole size to be.

    Related Pages:
http://nvd3.org/examples/pie.html

    Attributes:
    ----------
    :param flag:
    """
    self.donut(True)
    self.fnc("donutRatio(%s)" % JsUtils.jsConvertData(value, None))
    return self

  def padAngle(self, val):
    """
    Description:
    ------------
    Add a padding (space) between the different categories of the chart

    Related Pages:

			https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/donutChart.html

    Attributes:
    ----------
    :param val: Float. The padding value
    """
    self.fnc("padAngle(%s)" % JsUtils.jsConvertData(val, None))
    return self

  def cornerRadius(self, val):
    """
    Description:
    ------------
    Change the angle corner radius

    Related Pages:

			https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/donutChart.html

    Attributes:
    ----------
    :param val: Float. The rounding to be set to the angles
    """
    self.fnc("cornerRadius(%s)" % JsUtils.jsConvertData(val, None))
    return self

  def id(self, classname):
    """
    Description:
    ------------
    allow custom CSS for this one svg

    Attributes:
    ----------
    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/donutChart.html

    Attributes:
    ----------
    :param classname:
    """
    self.fnc("id(%s)" % JsUtils.jsConvertData(classname, None))
    return self

  def growOnHover(self, flag):
    """

    :param flag:
    """
    self.fnc("growOnHover(%s)" % JsUtils.jsConvertData(flag, None))
    return self

  def reduceXTicks(self, flag):
    pass

  def staggerLabels(self, flag):
    pass

  def tooltips(self, flag):
    pass

  def showValues(self, flag):
    pass

  def groupSpacing(self, value):
    pass

  def title(self, text):
    """
    Text to include within the middle of a donut chart

    Related Pages:

			https://nvd3-community.github.io/nvd3/examples/documentation.html

    :param text:

    :return:
    """
    self.fnc("title(%s)" % JsUtils.jsConvertData(JsUtils.jsConvertData(text, None), None))
    return self

  def titleOffset(self, value):
    pass


class JsNvd3ForceDirectedGraph(JsNvd3Bar):
  chartFnc = "forceDirectedGraph"

  def color(self, column):
    """

    :param column:
    """
    self.fnc("color(function(d) { return d.%s })" % column)
    return self

  def nodeExtras(self, text):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/forceDirected.html

    :param node:
    """

    self.fnc("nodeExtras(function(node) {node.append('text').attr('dx', 12).attr('dy', '.35em').text(function(d) { return d.%s }) })" % text)
    return self



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_punctuation.asp
"""

EN_QUAD = '&#8192'
EM_QUAD = '&#8193'
EN_SPACE = '&#8194'
EM_SPACE = '&#8195'
THREE_PER_EM_SPACE = '&#8196'
FOUR_PER_EM_SPACE = '&#8197'
SIX_PER_EM_SPACE = '&#8198'
FIGURE_SPACE = '&#8199'
PUNCTUATION_SPACE = '&#8200'
THIN_SPACE = '&#8201'
HAIR_SPACE = '&#8202'
ZERO_WIDTH_SPACE = '&#8203'
ZERO_WIDTH_NON_JOINER = '&#8204'
ZERO_WIDTH_JOINER = '&#8205'
LEFT_TO_RIGHT_MARK = '&#8206'
RIGHT_TO_LEFT_MARK = '&#8207'
HYPHEN = '&#8208'
NON_BREAKING_HYPHEN = '&#8209'
FIGURE_DASH = '&#8210'
EN_DASH = '&#8211'
EM_DASH = '&#8212'
HORIZONTAL_BAR = '&#8213'
DOUBLE_VERTICAL_LINE = '&#8214'
DOUBLE_LOW_LINE = '&#8215'
LEFT_SINGLE_QUOTATION_MARK = '&#8216'
RIGHT_SINGLE_QUOTATION_MARK = '&#8217'
SINGLE_LOW_9_QUOTATION_MARK = '&#8218'
SINGLE_HIGH_REVERSED_9_QUOTATION_MARK = '&#8219'
LEFT_DOUBLE_QUOTATION_MARK = '&#8220'
RIGHT_DOUBLE_QUOTATION_MARK = '&#8221'
DOUBLE_LOW_9_QUOTATION_MARK = '&#8222'
DOUBLE_HIGH_REVERSED_9_QUOTATION_MARK = '&#8223'
DAGGER = '&#8224'
DOUBLE_DAGGER = '&#8225'
BULLET = '&#8226'
TRIANGULAR_BULLET = '&#8227'
ONE_DOT_LEADER = '&#8228'
TWO_DOT_LEADER = '&#8229'
HORIZONTAL_ELLIPSIS = '&#8230'
HYPHENATION_POINT = '&#8231'
LINE_SEPARATOR = '&#8232'
PARAGRAPH_SEPARATOR = '&#8233'
LEFT_TO_RIGHT_EMBEDDING = '&#8234'
RIGHT_TO_LEFT_EMBEDDING = '&#8235'
POP_DIRECTIONAL_FORMATTING = '&#8236'
LEFT_TO_RIGHT_OVERRIDE = '&#8237'
RIGHT_TO_LEFT_OVERRIDE = '&#8238'
NARROW_NON_BREAK_SPACE = '&#8239'
PER_MILLE_SIGN = '&#8240'
PER_TEN_THOUSAND_SIGN = '&#8241'
PRIME = '&#8242'
DOUBLE_PRIME = '&#8243'
TRIPLE_PRIME = '&#8244'
REVERSED_PRIME = '&#8245'
REVERSED_DOUBLE_PRIME = '&#8246'
REVERSED_TRIPLE_PRIME = '&#8247'
CARET = '&#8248'
SINGLE_LEFT_POINTING_ANGLE_QUOTATION_MARK = '&#8249'
SINGLE_RIGHT_POINTING_ANGLE_QUOTATION_MARK = '&#8250'
REFERENCE_MARK = '&#8251'
DOUBLE_EXCLAMATION_MARK = '&#8252'
INTERROBANG = '&#8253'
OVERLINE = '&#8254'
UNDERTIE = '&#8255'
CHARACTER_TIE = '&#8256'
CARET_INSERTION_POINT = '&#8257'
ASTERISM = '&#8258'
HYPHEN_BULLET = '&#8259'
FRACTION_SLASH = '&#8260'
LEFT_SQUARE_BRACKET_WITH_QUILL = '&#8261'
RIGHT_SQUARE_BRACKET_WITH_QUILL = '&#8262'
DOUBLE_QUESTION_MARK = '&#8263'
QUESTION_EXCLAMATION_MARK = '&#8264'
EXCLAMATION_QUESTION_MARK = '&#8265'
TIRONIAN_SIGN_ET = '&#8266'
REVERSED_PILCROW_SIGN = '&#8267'
BLACK_LEFTWARDS_BULLET = '&#8268'
BLACK_RIGHTWARDS_BULLET = '&#8269'
LOW_ASTERISK = '&#8270'
REVERSED_SEMICOLON = '&#8271'
CLOSE_UP = '&#8272'
TWO_ASTERISKS_ALIGNED_VERTICALLY = '&#8273'
COMMERCIAL_MINUS_SIGN = '&#8274'
SWUNG_DASH = '&#8275'
INVERTED_UNDERTIE = '&#8276'
FLOWER_PUNCTUATION_MARK = '&#8277'
THREE_DOT_PUNCTUATION = '&#8278'
QUADRUPLE_PRIME = '&#8279'
FOUR_DOT_PUNCTUATION = '&#8280'
FIVE_DOT_PUNCTUATION = '&#8281'
TWO_DOT_PUNCTUATION = '&#8282'
FOUR_DOT_MARK = '&#8283'
DOTTED_CROSS = '&#8284'
TRICOLON = '&#8285'
VERTICAL_FOUR_DOTS = '&#8286'
MEDIUM_MATHEMATICAL_SPACE = '&#8287'
WORD_JOINER = '&#8288'
FUNCTION_APPLICATION = '&#8289'
INVISIBLE_TIMES = '&#8290'
INVISIBLE_SEPARATOR = '&#8291'
INVISIBLE_PLUS = '&#8292'
LEFT_TO_RIGHT_ISOLATE = '&#8294'
RIGHT_TO_LEFT_ISOLATE = '&#8295'
FIRST_STRONG_ISOLATE = '&#8296'
POP_DIRECTIONAL_ISOLATE = '&#8297'
INHIBIT_SYMMETRIC_SWAPPING = '&#8298'
ACTIVATE_SYMMETRIC_SWAPPING = '&#8299'
INHIBIT_ARABIC_FORM_SHAPING = '&#8300'
ACTIVATE_ARABIC_FORM_SHAPING = '&#8301'
NATIONAL_DIGIT_SHAPES = '&#8302'
NOMINAL_DIGIT_SHAPES = '&#8303'




import datetime

from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtmlEditor
from epyk.core.js.packages import JsCodeMirror
from epyk.core.html import Html

from epyk.core.html.options import OptCodeMirror
from epyk.core.html.options import OptText


class Console(Html.Html):
  name, category = 'Console', 'Rich'

  def __init__(self, report, data, width, height, htmlCode, helper, options, profile):
    super(Console, self).__init__(report, data, code=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.css({"overflow": 'auto', 'box-sizing': 'border-box', 'color': self._report.theme.greys[-1],
              'background': self._report.theme.colors[0]})
    self.__options = OptText.OptionsConsole(self, options)
    self.add_helper(helper)

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlEditor.Console
    """
    if self._dom is None:
      self._dom = JsHtmlEditor.Console(self, report=self._report)
    return self._dom

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptText.OptionsConsole
    """
    return self.__options

  @property
  def _js__builder__(self):
    return ''' 
      if(options.showdown){var converter = new showdown.Converter(options.showdown);
        let frag = document.createRange().createContextualFragment(converter.makeHtml(data)); 
        frag.firstChild.style.display = 'inline-block'; frag.firstChild.style.margin = 0;  
        data = frag.firstChild.outerHTML} 
      htmlObj.innerHTML = data +'<br/>' '''

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "<div %s></div>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Editor(Html.Html):
  name, category, callFnc = 'Code Editor', 'Text', 'editor'
  __reqCss, __reqJs = ['codemirror', 'font-awesome'], ['codemirror', 'font-awesome']

  def __init__(self, report, vals, language, width, height, htmlCode, options, profile):
    super(Editor, self).__init__(report, vals, code=htmlCode, css_attrs={"width": width, "height": height,
            'box-sizing': 'border-box', 'margin': '5px 0'}, profile=profile)
    self.textarea = self._report.ui.texts.code(vals, height=height, language=language, options=options)
    self.textarea.inReport = False
    self.actions = []

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlEditor.Editor
    """
    if self._dom is None:
      self._dom = JsHtmlEditor.Editor(self, report=self._report)
    return self._dom

  def action(self, icon, jsFncs, tooltip=None):
    """
    Description:
    ------------
    Add a bespoke action to the action panel

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover

    :return:
    """
    icon_button = self._report.ui.icon(icon, tooltip=tooltip).css({"margin-right": '5px'}).click(jsFncs)
    self.actions.append((icon, icon_button))
    icon_button.inReport = False
    return self

  def toggle(self, jsFncs, icons=("fas fa-eye", "far fa-eye-slash"), tooltip=None):
    """
    Description:
    ------------
    Add an event action to the console object.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover

    :return:
    """
    icon_button = self._report.ui.icon(icons[0], tooltip=tooltip).css({"margin-right": '5px'})
    jsFncs.append(self.textarea.dom.toggle())
    jsFncs.append(icon_button.dom.switchClass(icons[0], icons[1]).r)
    icon_button.click(jsFncs)
    icon_button.inReport = False
    self.actions.append((icons[0], icon_button))
    return self

  def copy(self, jsFncs, icon="far fa-clipboard", tooltip=None):
    """
    Description:
    ------------
    Copy the content of the editor component to the clipboard.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    jsFncs.append(self.textarea.dom.select())
    jsFncs.append('document.execCommand("copy")')
    return self.action(icon, jsFncs, tooltip)

  def run(self, jsFncs, icon="fas fa-play", tooltip=None):
    """
    Description:
    ------------
    Emtpy run button.
    This function will just add the icon on the actions panel.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    return self.action(icon, jsFncs, tooltip)

  def save(self, jsFncs, icon="fas fa-save", tooltip=None):
    """
    Description:
    ------------
    Emtpy save button.
    This function will just add the icon on the actions panel.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    return self.action(icon, jsFncs, tooltip)

  def clear(self, jsFncs, icon="fas fa-times-circle", tooltip=None):
    """
    Description:
    ------------
    Add an event action to the console object.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    jsFncs.append(self.textarea.dom.clear())
    return self.action(icon, jsFncs, tooltip)

  def __str__(self):
    actions = "".join([b.html() for _, b in self.actions])
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return '''
        <div %(attr)s>%(actions)s
          <span style="display:inline-block;float:right;margin-right:5px;font-style:italic">%(timestamp)s</span>
        </div> 
        %(textarea)s''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'timestamp': timestamp,
                           "textarea": self.textarea.html(), 'htmlId': self.htmlId, 'actions': actions}


class Cell(Html.Html):
  name, category, callFnc = 'Python Cell Runner', 'Text', 'pytestcell'
  __reqCss, __reqJs = ['codemirror'], ['codemirror']

  def __init__(self, report, vals, language, width, height, htmlCode, options, profile):
    super(Cell, self).__init__(report, vals, code=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.textarea = self._report.ui.texts.code(vals, height=height, language=language, options=options)
    self.textarea.inReport = False
    self.textarea.style.add_classes.input.textarea()
    self._jsRun, self._jsSave = '', ''
    self.css({'padding': '10px', "min-height": "30px", 'box-sizing': 'border-box', 'display': 'inline-block'})
    self.actions = []

  def action(self, icon, jsFncs, tooltip=None):
    """
    Description:
    ------------
    Add a bespoke action to the action panel

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover

    :return:
    """
    icon_button = self._report.ui.icon(icon, tooltip=tooltip).css({"margin-right": '5px'}).click(jsFncs)
    self.actions.append((icon, icon_button))
    icon_button.inReport = False

  def run(self, jsFncs, icon="fas fa-play", tooltip=None):
    """
    Description:
    ------------
    Emtpy run button.
    This function will just add the icon on the actions panel.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    jsFncs.append(self.dom.querySelector("span").innerHTML(1, append=True, valType=int))
    return self.action(icon, jsFncs, tooltip)

  def save(self, jsFncs, icon="fas fa-save", tooltip=None):
    """
    Description:
    ------------
    Emtpy save button.
    This function will just add the icon on the actions panel.

    Attributes:
    ----------
    :param icon: String. The font awesome icon
    :param jsFncs: Array. The Javascript functions
    :param tooltip: String. Text to be displayed when mouse is hover
    """
    return self.action(icon, jsFncs, tooltip)

  def __str__(self):
    actions = "".join([b.html() for _, b in self.actions])
    return '''
      <div %(attrs)s>
          <div style="padding:10px 5px;float:left;width:50px;height:100%%;vertical-align:middle">
            In [ <span data=count=0 style="display:inline-block;margin-bottom:5px">0</span> ]<br/>%(actions)s
          </div>
          %(textarea)s
      </div>''' % {'attrs': self.get_attrs(pyClassNames=self.style.get_classes()), 'actions': actions, "textarea": self.textarea.html()}


class Code(Html.Html):
  name, category, callFnc = 'Code', 'Text', 'code'
  __reqCss, __reqJs = ['codemirror'], ['codemirror']

  def __init__(self, report, vals, color, width, height, htmlCode, options, helper, profile):
    super(Code, self).__init__(report, vals, code=htmlCode, css_attrs={"width": width, "height": height, "color": color}, profile=profile)
    self.add_helper(helper)
    self.__options = OptCodeMirror.OptionsCode(self, options)
    self.css({'display': 'block', 'margin': '5px 0'})

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptCodeMirror.OptionsCode
    """
    return self.__options

  @property
  def js(self):
    """
    Description:
    -----------
    A lot of CodeMirror features are only available through its API. Thus, you need to write code (or use addons) if you want to expose them to your users.

    Related Pages:

			https://codemirror.net/doc/manual.html#api

    :rtype: JsCodeMirror.CM
    """
    if self._js is None:
      self._js = JsCodeMirror.CM(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlEditor.CodeMirror
    """
    if self._dom is None:
      self._dom = JsHtmlEditor.CodeMirror(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return '''
       htmlObj.setValue(data); Object.keys(options).forEach(function(key){ htmlObj.setOption(key, options[key])}); 
       htmlObj.refresh()'''

  def build(self, data=None, options=None, profile=False):
    if not self.builder_name:
      raise Exception("No builder defined for this HTML component %s" % self.__class__.__name__)

    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})

    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    options, js_options = options or self._jsStyles, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    #
    constructors[
      self.builder_name] = "var %(editorId)s = CodeMirror.fromTextArea(%(htmlId)s, {%(options)s}); %(editorId)s.setSize(null, '%(height)s'); function %(name)s(htmlObj, data, options){%(builder)s}" % {"editorId": self.editorId,
        'htmlId': self.htmlId, 'options': ",".join(js_options), 'name': self.builder_name, 'builder': self._js__builder__, 'height': self.attr['css']['height']}
    return "%s(%s, %s, %s)" % (self.builder_name, self.editorId, js_data, "{%s}" % ",".join(js_options))

  @property
  def editorId(self):
    """
    Description:
    ------------
    Return the Javascript variable of the bespoke
    """
    return "editor_%s" % self.htmlId

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<textarea %s></textarea>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Tags(Html.Html):
  name, category, callFnc = 'Tags', None, 'tags'
  # _grpCls = GrpCls.CssGrpClassBase

  def __init__(self, report, vals, title, icon, size, width, height, htmlCode, profile):
    super(Tags, self).__init__(report, vals, width=width[0], widthUnit=width[1], height=height[0], heightUnit=height[1],
                               code=htmlCode, profile=profile)
    self.title, self.icon = title, icon
    self.css({"margin-top": "5px", "font-size": "%s%s" % (size[0], size[1]), "font-family": report.style.defaults.font.family})

  @property
  def val(self):
    return "%(breadCrumVar)s['params']['%(htmlId)s']" % {"htmlId": self.htmlId, "breadCrumVar": self._report.jsGlobal.breadCrumVar}

  def jsEmpty(self):
    return "%(breadCrumVar)s['params']['%(htmlId)s'] = []; $('#%(htmlId)s_tags').text('')" % {"htmlId": self.htmlId, "breadCrumVar": self._report.jsGlobal.breadCrumVar}

  def jsAdd(self, jsData):
    jsData = JsUtils.jsConvertData(jsData, None)
    self.addGlobalFnc('RemoveSelection(srcObj, htmlId)', 'srcObj.parent().remove()',
       fncDsc="Remove the item from the Tags Html component but also from the underlying javascript variable")
    return '''
      $('#%(htmlId)s_tags').append("<span style='margin:2px;background:%(baseColor)s;color:%(whiteColor)s;border-radius:8px;1em;vertical-align:middle;display:inline-block;padding:0 2px 1px 10px;cursor:pointer'>"+ %(jsData)s +"<i onclick='RemoveSelection($(this), \\\"%(htmlId)s\\\")' style='margin-left:10px' class='far fa-times-circle'></i></span>")
      ''' % {"htmlId": self.htmlId, "jsData": jsData, 'whiteColor': self._report.theme.greys[0], "baseColor": self._report.theme.colors[9]}

  def __str__(self):
    return '''
      <div %(attr)s>
        <div style='margin:0;display:inline-block;vertical-align:middle;width:90px;float:left;padding:2px 5px 0 5px;height:30px;border:1px solid %(greyColor)s'>
          <i class="%(icon)s" style="margin-right:10px"></i>%(title)s</div>
        <div id='%(htmlId)s_tags' style='padding:2px 5px 0 5px;border:1px solid %(greyColor)s;height:30px'></div>
      </div>''' % {"attr": self.get_attrs(pyClassNames=self.defined), "title": self.title, 'icon': self.icon,
                   'htmlId': self.htmlId, 'greyColor': self._report.theme.greys[2]}
